import torch.nn as nn
import torch
import torch.nn.functional as F
from . import block as B
from .bam import *


"""
SR-ResNet Architecture with Continuous Attention to Candidates
"""
class Conditioned_SRResNet_WithCand_Attenv2(nn.Module):
    # in_nc=1, out_nc=1, nf=64, nb=8, upscale=1, norm_type=None, act_type='relu' 
    def __init__(self, in_nc, out_nc, nf, nb, kernel_class, dose_class, upscale=1, norm_type='batch', act_type='relu'):
        super(Conditioned_SRResNet_WithCand_Attenv2, self).__init__()

        # label embeddings
        self.gen_kernel_embedding = nn.Sequential(nn.Embedding(kernel_class, kernel_class), nn.Linear(kernel_class, 32*64*64))
        self.gen_dose_embedding = nn.Sequential(nn.Embedding(dose_class, dose_class), nn.Linear(dose_class, 32*64*64))

        # head-block
        """
        Conv3d(1, 64, kernel_size=(3, 3, 3), stride=(1, 1, 1), padding=(1, 1, 1))
        """
        self.head_conv = nn.Conv3d(in_nc+2, nf, 3, 1, 1, bias=True)
        self.merge_cand = nn.quantized.FloatFunctional()
        
        # body-block: create resnet blocks depending on 'nb' parameters, with attention modules
        attention_blocks = [BAM(nf) for _ in range(nb)]
        self.attn_blocks = nn.Sequential(*attention_blocks)
        resnet_blocks = [ResNetBlock(nf, nf, kernel_size=3, act_type=act_type) for _ in range(nb)]
        # self.net = B.sequential(B.ShortcutBlock(B.sequential(*resnet_blocks, lr_conv)))
        self.net = nn.Sequential(*resnet_blocks)
        self.final_residual = B.conv_block(nf, nf, kernel_size=3, norm_type=norm_type, act_type=None)

        # tail-block
        self.hr_conv = B.conv_block(nf, nf, kernel_size=3, norm_type=norm_type, act_type=act_type)
        self.last_conv = nn.Conv3d(nf, out_nc, 3, 1, 1, bias=True)

    def forward(self, x, gen_kernel_lbl, gen_dose_lbl, candidates):
        # pass through label embeddings
        gen_kernel_embed = self.gen_kernel_embedding(gen_kernel_lbl).view(-1, x.shape[1], x.shape[2], x.shape[3], x.shape[4])
        gen_dose_embed = self.gen_dose_embedding(gen_dose_lbl).view(-1, x.shape[1], x.shape[2], x.shape[3], x.shape[4])
        # add embedding to image and candidate
        x = torch.cat((x, gen_kernel_embed, gen_dose_embed), dim=1)
        x_cand = torch.cat((candidates, gen_kernel_embed, gen_dose_embed), dim=1)
        # pass image and candidate through head-block
        x = self.head_conv(x)
        reuse_head = x.clone() # resuse after residual blocks
        x_cand = self.head_conv(x_cand)
        # x_with_cand = self.merge_cand.add(x, x_cand) # add image with candidates
        # pass through body block
        for _ in range(len(self.net)):
            x_cand = self.attn_blocks[_](x_cand)
            x = self.net[_](x)
            x = self.merge_cand.add(x, x_cand)
        x = self.merge_cand.add(reuse_head, self.final_residual(x))
        # pass through tail block
        x = self.hr_conv(x)
        x = self.last_conv(x)
        return x


"""
SR-ResNet Architecture with Attention to Candidates
"""
class Conditioned_SRResNet_WithCand_Atten(nn.Module):
    # in_nc=1, out_nc=1, nf=64, nb=8, upscale=1, norm_type=None, act_type='relu' 
    def __init__(self, in_nc, out_nc, nf, nb, kernel_class, dose_class, upscale=1, norm_type='batch', act_type='relu'):
        super(Conditioned_SRResNet_WithCand_Atten, self).__init__()

        # label embeddings
        self.gen_kernel_embedding = nn.Sequential(nn.Embedding(kernel_class, kernel_class), nn.Linear(kernel_class, 32*64*64))
        self.gen_dose_embedding = nn.Sequential(nn.Embedding(dose_class, dose_class), nn.Linear(dose_class, 32*64*64))

        # head
        """
        Conv3d(1, 64, kernel_size=(3, 3, 3), stride=(1, 1, 1), padding=(1, 1, 1))
        """
        self.head_conv = nn.Conv3d(in_nc+2, nf, 3, 1, 1, bias=True)
        self.merge_cand = nn.quantized.FloatFunctional()
        
        # body: create a number of resnet blocks depending on 'nb' parameters, with attention modules
        attention = [BAM(nf) for _ in range(nb)]
        resnet_blocks = [ResNetBlockWithAtten(nf, nf, atten_module=attention[_], kernel_size=3, act_type=act_type) for _ in range(nb)]
        # body: after end of all residual blocks
        lr_conv = B.conv_block(nf, nf, kernel_size=3, norm_type=norm_type, act_type=None)

        # The 'up' and 'up_conv' are only used when 'upscale' is not 1. i.e used only for SR 
        # upsample (scale factor is Depth, Height, Width). For upsampling depth, factor comes from config file
        up = nn.Upsample(scale_factor=(upscale, 1., 1.,), mode='nearest')
        up_conv = B.conv_block(nf, nf, kernel_size=3, norm_type=norm_type, act_type=act_type)
        
        # tail
        hr_conv = B.conv_block(nf, nf, kernel_size=3, norm_type=norm_type, act_type=act_type)
        last_conv = nn.Conv3d(nf, out_nc, 3, 1, 1, bias=True)

        # unpack all above defined layers, e.g. 'fea_conv', 'hr_conv', ..., into one 'nn.Sequential' module
        if upscale == 1: # denoising
            self.net = B.sequential(B.ShortcutBlock(B.sequential(*resnet_blocks, lr_conv)), hr_conv, last_conv)
        else: # super-res
            self.net = B.sequential(B.ShortcutBlock(B.sequential(*resnet_blocks, lr_conv)),
                                    up, up_conv, hr_conv, last_conv)

    def forward(self, x, gen_kernel_lbl, gen_dose_lbl, candidates):
        gen_kernel_embed = self.gen_kernel_embedding(gen_kernel_lbl).view(-1, x.shape[1], x.shape[2], x.shape[3], x.shape[4])
        gen_dose_embed = self.gen_dose_embedding(gen_dose_lbl).view(-1, x.shape[1], x.shape[2], x.shape[3], x.shape[4])
        # add embedding to image and candidate
        x = torch.cat((x, gen_kernel_embed, gen_dose_embed), dim=1)
        x_cand = torch.cat((candidates, gen_kernel_embed, gen_dose_embed), dim=1)
        # pass through head
        x = self.head_conv(x)
        x_cand = self.head_conv(x_cand)
        x_with_can = self.merge_cand.add(x, x_cand)
        x = self.net(x_with_can)
        return x


"""
SR-ResNet Architecture with Label Encoding 
"""
class Conditioned_SRResNet(nn.Module):
    # in_nc=1, out_nc=1, nf=64, nb=8, upscale=1, norm_type=None, act_type='relu' 
    def __init__(self, in_nc, out_nc, nf, nb, kernel_class, dose_class, upscale=1, norm_type='batch', act_type='relu', atten_lyrs=False):
        super(Conditioned_SRResNet, self).__init__()

        # label embeddings
        self.gen_kernel_embedding = nn.Sequential(nn.Embedding(kernel_class, kernel_class), nn.Linear(kernel_class, 32*64*64))
        self.gen_dose_embedding = nn.Sequential(nn.Embedding(dose_class, dose_class), nn.Linear(dose_class, 32*64*64))

        # head
        """
        Conv3d(1, 64, kernel_size=(3, 3, 3), stride=(1, 1, 1), padding=(1, 1, 1))
        """
        fea_conv = nn.Conv3d(in_nc+2, nf, 3, 1, 1, bias=True)
        
        # body: create a number of resnet blocks depending on 'nb' parameters
        if atten_lyrs:
            attention = [BAM(nf) for _ in range(nb)]
            resnet_blocks = [ResNetBlockWithAtten(nf, nf, atten_module=attention[_], kernel_size=3, act_type=act_type) for _ in range(nb)]
        else:
            resnet_blocks = [ResNetBlock(nf, nf, kernel_size=3, act_type=act_type) for _ in range(nb)]
        # body: after end of all residual blocks
        lr_conv = B.conv_block(nf, nf, kernel_size=3, norm_type=norm_type, act_type=None)

        # The 'up' and 'up_conv' are only used when 'upscale' is not 1. i.e used only for SR 
        # upsample (scale factor is Depth, Height, Width). For upsampling depth, factor comes from config file
        up = nn.Upsample(scale_factor=(upscale, 1., 1.,), mode='nearest')
        # print("up:", up)
        up_conv = B.conv_block(nf, nf, kernel_size=3, norm_type=norm_type, act_type=act_type)
        # print("up-conv:", up_conv)
        
        # tail
        hr_conv = B.conv_block(nf, nf, kernel_size=3, norm_type=norm_type, act_type=act_type)
        # print("hr conv:", hr_conv)
        last_conv = nn.Conv3d(nf, out_nc, 3, 1, 1, bias=True)
        # print("last conv:", last_conv)

        # unpack all above defined layers, e.g. 'fea_conv', 'hr_conv', ..., into one 'nn.Sequential' module
        if upscale == 1: # denoising
            self.net = B.sequential(fea_conv, B.ShortcutBlock(B.sequential(*resnet_blocks, lr_conv)), hr_conv, last_conv)
        else: # super-res
            self.net = B.sequential(fea_conv, B.ShortcutBlock(B.sequential(*resnet_blocks, lr_conv)),
                                    up, up_conv, hr_conv, last_conv)

    def forward(self, x, gen_kernel_lbl, gen_dose_lbl):
        gen_kernel_embed = self.gen_kernel_embedding(gen_kernel_lbl).view(-1, x.shape[1], x.shape[2], x.shape[3], x.shape[4])
        gen_dose_embed = self.gen_dose_embedding(gen_dose_lbl).view(-1, x.shape[1], x.shape[2], x.shape[3], x.shape[4])
        x = torch.cat((x, gen_kernel_embed, gen_dose_embed), dim=1)
        x = self.net(x)
        return x


"""
SR-ResNet Architecture
"""
class SRResNet(nn.Module):
    # in_nc=1, out_nc=1, nf=64, nb=8, upscale=1, norm_type=None, act_type='relu' 
    def __init__(self, in_nc, out_nc, nf, nb, upscale=1, norm_type='batch', act_type='relu'):
        super(SRResNet, self).__init__()
        # head
        """
        Conv3d(1, 64, kernel_size=(3, 3, 3), stride=(1, 1, 1), padding=(1, 1, 1))
        """
        fea_conv = nn.Conv3d(in_nc, nf, 3, 1, 1, bias=True)
        # print("feature 1:", fea_conv)

        # body: create a number of resnet blocks depending on 'nb' parameters
        resnet_blocks = [ResNetBlock(nf, nf, 3, act_type=act_type) for _ in range(nb)]
        # body: after end of all residual blocks
        lr_conv = B.conv_block(nf, nf, kernel_size=3, norm_type=norm_type, act_type=None)
        # print("lr conv:", lr_conv)

        # The 'up' and 'up_conv' are only used when 'upscale' is not 1. i.e used only for SR 
        # upsample (scale factor is Depth, Height, Width). For upsampling depth, factor comes from config file
        up = nn.Upsample(scale_factor=(upscale, 1., 1.,), mode='nearest')
        # print("up:", up)
        up_conv = B.conv_block(nf, nf, kernel_size=3, norm_type=norm_type, act_type=act_type)
        # print("up-conv:", up_conv)

        # tail
        hr_conv = B.conv_block(nf, nf, kernel_size=3, norm_type=norm_type, act_type=act_type)
        # print("hr conv:", hr_conv)
        last_conv = nn.Conv3d(nf, out_nc, 3, 1, 1, bias=True)
        # print("last conv:", last_conv)

        # unpack all above defined layers, e.g. 'fea_conv', 'hr_conv', ..., into one 'nn.Sequential' module
        if upscale == 1:# denoising
            self.net = B.sequential(fea_conv, B.ShortcutBlock(B.sequential(*resnet_blocks, lr_conv)),
                                    hr_conv, last_conv)
        else: # super-res
            self.net = B.sequential(fea_conv, B.ShortcutBlock(B.sequential(*resnet_blocks, lr_conv)),
                                    up, up_conv, hr_conv, last_conv)
        """
        # initiazlie kaiming normal
        for key in self.state_dict():
            if key.split('.')[-1]=="weight":
                if "conv" in key:
                    # init.kaiming_normal(self.state_dict()[key], mode='fan_out')
                    init.kaiming_normal_(self.state_dict()[key], mode='fan_out')
                if "bn" in key:
                    if "SpatialGate" in key:
                        self.state_dict()[key][...] = 0
                    else:
                        self.state_dict()[key][...] = 1
            elif key.split(".")[-1]=='bias':
                self.state_dict()[key][...] = 0
        """

    def forward(self, x):
        return self.net(x)


class QuantizedModel(nn.Module):
    # https://leimao.github.io/blog/PyTorch-Static-Quantization/
    def __init__(self, model_fp32):
        super(QuantizedModel, self).__init__()
        # QuantStub converts tensors from floating point to quantized.
        # This will only be used for inputs.
        self.quant = torch.quantization.QuantStub()
        # DeQuantStub converts tensors from quantized to floating point.
        # This will only be used for outputs.
        self.dequant = torch.quantization.DeQuantStub()
        # FP32 model
        self.model_fp32 = model_fp32

    def forward(self, x):
        # manually specify where tensors will be converted from floating
        # point to quantized in the quantized model
        x = self.quant(x)
        x = self.model_fp32(x)
        # manually specify where tensors will be converted from quantized
        # to floating point in the quantized model
        x = self.dequant(x)
        return x


# WGAN paper  https://arxiv.org/abs/1708.00961 only for denoising
# Low Dose CT Image Denoising Using a Generative Adversarial Network with Wasserstein Distance
# and Perceptual Loss
class VanillaNet(nn.Module):
    def __init__(self, in_nc, out_nc, nf, nb):
        super(VanillaNet, self).__init__()
        layers = [nn.Conv3d(in_nc, nf, 3, 1, 1), nn.ReLU()]
        for _ in range(2, nb):
            layers.extend([nn.Conv3d(nf, nf, 3, 1, 1), nn.ReLU()])
        layers.extend([nn.Conv3d(nf, out_nc, 3, 1, 1)]) # removed ReLU layer for stability
        self.net = nn.Sequential(*layers)

    def forward(self, x):
        out = self.net(x)
        return out


class ResNetBlock(nn.Module):
    '''
    ResNet Block, 3-3 style
    with extra residual scaling used in EDSR
    (Enhanced Deep Residual Networks for Single Image Super-Resolution, CVPRW 17)
    '''
    # in_nc=64, out_nc=64, act_type='relu', norm_type=None 
    def __init__(self, in_nc, out_nc, kernel_size=3, stride=1, dilation=1, bias=True, \
                 norm_type=None, act_type='relu', res_scale=1):
        super(ResNetBlock, self).__init__()
        """
        conv0 -> Sequential(
        (0): Conv3d(64, 64, kernel_size=(3, 3, 3), stride=(1, 1, 1), padding=(1, 1, 1))
        (1): ReLU(inplace=True)
        )
        """
        # has activation layer
        conv0 = B.conv_block(in_nc, out_nc, kernel_size, stride, dilation, bias, norm_type, act_type)
        """
        conv1 -> Sequential(
        (0): Conv3d(64, 64, kernel_size=(3, 3, 3), stride=(1, 1, 1), padding=(1, 1, 1))
        )
        """
        # does not have activation layer
        conv1 = B.conv_block(out_nc, out_nc, kernel_size, stride, dilation, bias, norm_type, None)
        """
        self.res -> Sequential(
        (0): Conv3d(64, 64, kernel_size=(3, 3, 3), stride=(1, 1, 1), padding=(1, 1, 1))
        (1): ReLU(inplace=True)
        (2): Conv3d(64, 64, kernel_size=(3, 3, 3), stride=(1, 1, 1), padding=(1, 1, 1))
        )
        """
        self.res = B.sequential(conv0, conv1)
        self.res_scale = res_scale
        """
        Use 'torch.nn.quantized.FloatFunctional' to wrap tensor operations that require special handling 
        for quantization into modules. 
        - Examples are operations like 'add' and 'cat' which require special handling to determine 
        output quantization parameters.
        """
        self.skip_op = nn.quantized.FloatFunctional()
        # print("resnetblock:", self.res)
        # print("skip:", self.skip_op)


    def forward(self, x):
        # multiply residual of x 'self.res(x)' with a scalar factor 'self.res_scale'
        res = self.skip_op.mul_scalar(self.res(x), self.res_scale)
        # return addition of original input 'x' with its residual 'res'
        return self.skip_op.add(x, res)


"""
class ResNetBlockWithCandAtten(nn.Module):
    # in_nc=64, out_nc=64, act_type='relu', norm_type=None 
    def __init__(self, in_nc, out_nc, atten_module, kernel_size=3, stride=1, dilation=1, bias=True, \
                 norm_type=None, act_type='relu', res_scale=1):
        super(ResNetBlockWithCandAtten, self).__init__()
        # has activation layer
        conv0 = B.conv_block(in_nc, out_nc, kernel_size, stride, dilation, bias, norm_type, act_type)
        # does not have activation layer
        conv1 = B.conv_block(out_nc, out_nc, kernel_size, stride, dilation, bias, norm_type, None)
        self.res = B.sequential(conv0, conv1)
        self.res_scale = res_scale
        self.skip_op = nn.quantized.FloatFunctional()
        self.atten_module = atten_module

    def forward(self, x, x_candidate):
        res = self.skip_op.mul_scalar(self.res(x), self.res_scale)
        res = self.skip_op.add(x, res)
        res_with_atten = self.atten_module(res)
        return res_with_atten
"""


class ResNetBlockWithAtten(nn.Module):
    # in_nc=64, out_nc=64, act_type='relu', norm_type=None 
    def __init__(self, in_nc, out_nc, atten_module, kernel_size=3, stride=1, dilation=1, bias=True, \
                 norm_type=None, act_type='relu', res_scale=1):
        super(ResNetBlockWithAtten, self).__init__()
        # has activation layer
        conv0 = B.conv_block(in_nc, out_nc, kernel_size, stride, dilation, bias, norm_type, act_type)
        # does not have activation layer
        conv1 = B.conv_block(out_nc, out_nc, kernel_size, stride, dilation, bias, norm_type, None)
        self.res = B.sequential(conv0, conv1)
        self.res_scale = res_scale
        self.skip_op = nn.quantized.FloatFunctional()
        self.atten_module = atten_module

    def forward(self, x):
        res = self.skip_op.mul_scalar(self.res(x), self.res_scale)
        res = self.skip_op.add(x, res)
        res_with_atten = self.atten_module(res)
        return res_with_atten


####################
# FSRCNN
####################
class FSRCNN(nn.Module):
    def __init__(self, scale_factor=1, num_channels=1, d=56, s=12, m=4):
        super(FSRCNN, self).__init__()
        self.first_part = nn.Sequential(
            nn.Conv3d(num_channels, d, kernel_size=5, padding=5//2),
            nn.PReLU(d)
        )
        self.mid_part = [nn.Conv3d(d, s, kernel_size=1), nn.PReLU(s)]
        for _ in range(m):
            self.mid_part.extend([nn.Conv3d(s, s, kernel_size=3, padding=3//2), nn.PReLU(s)])
        self.mid_part.extend([nn.Conv3d(s, d, kernel_size=1), nn.PReLU(d)])
        self.mid_part = nn.Sequential(*self.mid_part)
        self.last_part = nn.ConvTranspose3d(d, num_channels, kernel_size=(9, 5, 5), stride=(scale_factor, 1, 1), padding=(4, 2, 2),
                                            output_padding=(scale_factor-1, 0, 0))

    def forward(self, x):
        x = self.first_part(x)
        x = self.mid_part(x)
        return self.last_part(x)


####################
# SLResNet
####################
class SLResNet(nn.Module):
    def __init__(self, in_nc, out_nc, nf, nb, inter_nc, upscale=1, norm_type='batch', act_type='relu'):
        super(SLResNet, self).__init__()

        # feature extraction/ denoise
        # fea_conv = B.conv_block(in_nc, nf, kernel_size=3, norm_type=norm_type, act_type=act_type)
        fea_conv = nn.Conv3d(in_nc, nf, kernel_size=5, padding=5//2, bias=True)
        resnet_blocks = [SL_A_ResNetBlock(nf, nf, inter_nc, 3, act_type=act_type) for _ in range(nb)]
        lr_conv = B.conv_block(nf, nf, kernel_size=3, norm_type=norm_type, act_type=None)
        # upsample 
        up = nn.Upsample(scale_factor=(upscale, 1., 1.,), mode='nearest')
        up_conv = B.conv_block(nf, nf, kernel_size=3, norm_type=norm_type, act_type=act_type)
        # output
        hr_conv = B.conv_block(nf, nf, kernel_size=3, norm_type=norm_type, act_type=act_type)
        conv_last = nn.Conv3d(nf, out_nc, 3, 1, 1, bias=True)

        if upscale == 1:# denoising
            self.net = B.sequential(fea_conv, B.ShortcutBlock(B.sequential(*resnet_blocks, lr_conv)),
                    hr_conv, conv_last)
        else: #super-res
            self.net = B.sequential(fea_conv, B.ShortcutBlock(B.sequential(*resnet_blocks, lr_conv)),
                                up, up_conv, hr_conv, conv_last)

    def forward(self, x):
        return self.net(x)

class SL_A_ResNetBlock(nn.Module):
    def __init__(self, in_nc, out_nc, inter_nc, kernel_size=3, stride=1, dilation=1, bias=True, \
                 norm_type=None, act_type='relu'):
        super(SL_A_ResNetBlock, self).__init__()
        spa_conv = B.s_conv_block(in_nc, inter_nc, kernel_size, stride, dilation, bias, norm_type, act_type)
        temp_conv = B.t_conv_block(inter_nc, out_nc, kernel_size, stride, dilation, bias, norm_type, None)
        self.res = B.sequential(spa_conv, temp_conv)
        self.skip_op = nn.quantized.FloatFunctional()

    def forward(self, x):
        return self.skip_op.add(x, self.res(x))

class SL_B_ResNetBlock(nn.Module):
    def __init__(self, in_nc, out_nc, bk_nc, inter_nc, kernel_size=3, stride=1, dilation=1, bias=True, \
                 norm_type=None, act_type='relu'):
        super(SL_B_ResNetBlock, self).__init__()

        self.spa_conv = B.s_conv_block(in_nc, inter_nc, kernel_size, stride, dilation, bias, norm_type, act_type)
        self.temp_conv = B.t_conv_block(in_nc, inter_nc, kernel_size, stride, dilation, bias, norm_type, act_type)
        self.conv1 = B.conv_block(inter_nc, out_nc, 1, stride, dilation, bias, norm_type, None)
        self.is_not_out_nc_equal_to_inter_nc = out_nc != inter_nc

    def forward(self, x):
        out = self.spa_conv(x) + self.temp_conv(x) # parallel spatial temporal
        if self.is_not_out_nc_equal_to_inter_nc:
            out = self.conv1(out) # 1x1x1 conv when out_nc != inter_nc
        return x + out

class SL_C_ResNetBlock(nn.Module):
    def __init__(self, in_nc, out_nc, bk_nc, inter_nc, kernel_size=3, stride=1, dilation=1, bias=True, \
                 norm_type=None, act_type='relu'):
        super(SL_C_ResNetBlock, self).__init__()

        self.spa_conv = B.s_conv_block(in_nc, inter_nc, kernel_size, stride, dilation, bias, norm_type, act_type)
        self.temp_conv = B.t_conv_block(inter_nc, inter_nc, kernel_size, stride, dilation, bias, norm_type, act_type)
        self.conv1 = B.conv_block(inter_nc, out_nc, 1, stride, dilation, bias, norm_type, None)

        self.is_not_out_nc_equal_to_inter_nc = out_nc != inter_nc

    def forward(self, x):
        out = self.spa_conv(x) 
        out = out + self.temp_conv(out) # fuse spatial & temporal
        if self.is_not_out_nc_equal_to_inter_nc:
            out = self.conv1(out) # 1x1x1 conv when out_nc != inter_nc
        return x + out


"""
Dense block for RRBD MODEL
"""
class ResidualDenseBlock(nn.Module):
    """Achieves densely connected convolutional layers.
    `Densely Connected Convolutional Networks <https://arxiv.org/pdf/1608.06993v5.pdf>` paper.

    Args:
        channels (int): The number of channels in the input image.
        growth_channels (int): The number of channels that increase in each layer of convolution.
    """
    def __init__(self, channels: int, growth_channels: int) -> None:
        super(ResidualDenseBlock, self).__init__()
        self.conv1 = nn.Conv3d(channels + growth_channels * 0, growth_channels, (3, 3, 3), (1, 1, 1), (1, 1, 1))
        self.conv2 = nn.Conv3d(channels + growth_channels * 1, growth_channels, (3, 3, 3), (1, 1, 1), (1, 1, 1))
        self.conv3 = nn.Conv3d(channels + growth_channels * 2, growth_channels, (3, 3, 3), (1, 1, 1), (1, 1, 1))
        self.conv4 = nn.Conv3d(channels + growth_channels * 3, growth_channels, (3, 3, 3), (1, 1, 1), (1, 1, 1))
        self.conv5 = nn.Conv3d(channels + growth_channels * 4, channels, (3, 3, 3), (1, 1, 1), (1, 1, 1))
        self.leaky_relu = nn.LeakyReLU(0.2, True)
        self.identity = nn.Identity()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x
        out1 = self.leaky_relu(self.conv1(x))
        out2 = self.leaky_relu(self.conv2(torch.cat([x, out1], 1)))
        out3 = self.leaky_relu(self.conv3(torch.cat([x, out1, out2], 1)))
        out4 = self.leaky_relu(self.conv4(torch.cat([x, out1, out2, out3], 1)))
        out5 = self.identity(self.conv5(torch.cat([x, out1, out2, out3, out4], 1)))
        out = torch.mul(out5, 0.2)
        out = torch.add(out, identity)
        return out


class ResidualResidualDenseBlock(nn.Module):
    """Multi-layer residual dense convolution block.
    Args:
        channels (int): The number of channels in the input image.
        growth_channels (int): The number of channels that increase in each layer of convolution.
    """
    def __init__(self, channels: int, growth_channels: int) -> None:
        super(ResidualResidualDenseBlock, self).__init__()
        self.rdb1 = ResidualDenseBlock(channels, growth_channels)
        self.rdb2 = ResidualDenseBlock(channels, growth_channels)
        self.rdb3 = ResidualDenseBlock(channels, growth_channels)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x
        out = self.rdb1(x)
        out = self.rdb2(out)
        out = self.rdb3(out)
        out = torch.mul(out, 0.2)
        out = torch.add(out, identity)
        return out


"""
RRBD MODEL
"""
class RRDB_Gen(nn.Module):
    def __init__(self, in_nc, out_nc, nf, nb, upscale=1):
        super(RRDB_Gen, self).__init__()
        self.conv1 = nn.Conv3d(in_nc, nf, (3, 3, 3), (1, 1, 1), (1, 1, 1))
        # Feature extraction backbone network
        trunk = []
        for _ in range(nb):
            trunk.append(ResidualResidualDenseBlock(nf, 32))
        self.trunk = nn.Sequential(*trunk)
        self.conv2 = nn.Conv3d(nf, nf, (3, 3, 3), (1, 1, 1), (1, 1, 1))
        # Reconnect a layer of convolution block after upsampling
        self.conv3 = nn.Sequential(
            nn.Conv3d(nf, nf, (3, 3, 3), (1, 1, 1), (1, 1, 1)),
            nn.LeakyReLU(0.2, True)
        )
        # Output layer
        self.conv4 = nn.Conv3d(nf, out_nc, (3, 3, 3), (1, 1, 1), (1, 1, 1))

    def _forward_impl(self, x: torch.Tensor) -> torch.Tensor:
        out1 = self.conv1(x)
        out = self.trunk(out1)
        out2 = self.conv2(out)
        out = torch.add(out1, out2)
        out = self.conv3(out)
        out = self.conv4(out)
        out = torch.clamp_(out, 0.0, 1.0)
        return out

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self._forward_impl(x)




"""
####################
# RRDB
####################
class RRDBNet(nn.Module):
    def __init__(self, in_nc, out_nc, nf, nb, gc=32, upscale=1, norm_type=None):
        super(RRDBNet, self).__init__()
        self.upscale = upscale

        # feature extraction/ denoise
        fea_conv = nn.Conv3d(in_nc, nf, 3, 1, 1, bias=True)
        rb_blocks = [RRDB(nf, gc) for _ in range(nb)]
        LR_conv = B.conv_block(nf, nf, kernel_size=3, norm_type=norm_type, act_type=None)
        self.out_feat = B.sequential(fea_conv, B.ShortcutBlock(B.sequential(*rb_blocks, LR_conv)))
        # upsample x2
        self.upconv = nn.Conv3d(nf, nf, 3, 1, 1, bias=True)
        self.lrelu = nn.LeakyReLU(negative_slope=0.2, inplace=True)
        # output
        self.HR_conv = nn.Conv3d(nf, nf, 3, 1, 1, bias=True)
        self.conv_last = nn.Conv3d(nf, out_nc, 3, 1, 1, bias=True)

    def forward(self, x):
        x = self.out_feat(x)
        if self.upscale == 2:
            x = self.lrelu(self.upconv(F.interpolate(x, scale_factor=(2., 1., 1.), mode='nearest')))
        # out = self.conv_last(x)
        return self.conv_last(self.lrelu(self.HR_conv(x)))


class ResidualDenseBlock_5C(nn.Module):
    '''
    Residual Dense Block
    style: 5 convs
    The core module of paper: (Residual Dense Network for Image Super-Resolution, CVPR 18)
    '''
    def __init__(self, nf, gc, kernel_size=3, stride=1, bias=True,
                 norm_type=None, act_type='leakyrelu'):
        super(ResidualDenseBlock_5C, self).__init__()
        # gc: growth channel, i.e. intermediate channels
        self.conv1 = B.conv_block(nf, gc, kernel_size, stride, bias=bias, \
            norm_type=norm_type, act_type=act_type)
        self.conv2 = B.conv_block(nf+gc, gc, kernel_size, stride, bias=bias, \
            norm_type=norm_type, act_type=act_type)
        self.conv3 = B.conv_block(nf+2*gc, gc, kernel_size, stride, bias=bias, \
            norm_type=norm_type, act_type=act_type)
        self.conv4 = B.conv_block(nf+3*gc, gc, kernel_size, stride, bias=bias, \
            norm_type=norm_type, act_type=act_type)
        self.conv5 = B.conv_block(nf+4*gc, nf, 3, stride, bias=bias, \
            norm_type=norm_type, act_type=None)

    def forward(self, x):
        x1 = self.conv1(x)
        x2 = self.conv2(torch.cat((x, x1), 1))
        x3 = self.conv3(torch.cat((x, x1, x2), 1))
        x4 = self.conv4(torch.cat((x, x1, x2, x3), 1))
        x5 = self.conv5(torch.cat((x, x1, x2, x3, x4), 1))
        return x5 * 0.2 + x


class RRDB(nn.Module):
    '''
    Residual in Residual Dense Block
    (ESRGAN: Enhanced Super-Resolution Generative Adversarial Networks)
    '''
    def __init__(self, nf=64, gc=32):
        super(RRDB, self).__init__()
        self.RDB1 = ResidualDenseBlock_5C(nf, gc)
        self.RDB2 = ResidualDenseBlock_5C(nf, gc)
        self.RDB3 = ResidualDenseBlock_5C(nf, gc)

    def forward(self, x):
        out = self.RDB1(x)
        out = self.RDB2(out)
        out = self.RDB3(out)
        return out * 0.2 + x
"""


import numpy as np
import h5py
import torch.utils.data as data
import utils.util as util
import random
import os
import torch


class h5Dataset(data.Dataset):
    def __init__(self, opt):
        super(h5Dataset, self).__init__()
        # useful when extracting 3D mask cube from the input masks
        self.FILL_RATIO_THRESHOLD = 0.8
        self.opt = opt # 'opt' is dictionary of parameters for either training/validation
        self.in_folder = opt['dataroot_LR']
        # print("-"*40)
        # print("input folder is:", self.in_folder)
        self.tar_folder = opt['dataroot_HR']
        # print("target folder is:", self.tar_folder)
        self.mask_folder = opt['maskroot_HR']
        # print("mask folder is:", self.mask_folder)
        # print("phase:", self.opt['phase'])

        # 3d voxel size
        if self.opt['phase'] == 'train' or self.opt['need_voxels']:
            self.ps = (opt['LR_slice_size'], opt['LR_size'], opt['LR_size'])
            # print("ps:", self.ps)        
        self.uids = opt['uids'] # list of uids
        # get only a subset of uids if provided
        if opt['subset'] is not None:
           self.uids = self.uids[:opt['subset']]
        self.scale = opt['scale']
        # print("scale:", self.scale)
        self.ToTensor = util.ImgToTensor()
        # print("-"*40)

    def __getitem__(self, index):
        uid = self.uids[index]
        # print("selected uid {}: {}".format(index, uid))
        # body mask - first we look at if the random voxel contain 80% of the body mask
        vol_mask = None
        if self.mask_folder:
            with h5py.File(os.path.join(self.mask_folder, uid+'.h5'), 'r') as file:
                IMG_THICKNESS, IMG_WIDTH, IMG_HEIGHT = file['data'].shape
                # random index of the 3D voxel for each patient
                if self.opt['phase'] == 'train' or (self.opt['need_voxels'] and not self.opt['need_voxels']['tile_x_y']):
                    t, w, h = self.ps
                    # print("t, w, h = {}, {}, {}".format(t, w, h))
                    # randomly search the voxel until 80% filled with ones
                    fill_ratio = 0.
                    """
                    randomly search for a 3D mask cube of size specified in t,w,h. 80% of the mask
                    must be filled with ones
                    """
                    # print("checking for fill ratio")
                    # print("*"*40)
                    while fill_ratio < self.FILL_RATIO_THRESHOLD:
                        rnd_t_HR = random.randint(0, IMG_THICKNESS - int(t * self.scale))
                        rnd_h = random.randint(0, IMG_HEIGHT - h)
                        rnd_w = random.randint(0, IMG_WIDTH - w)
                        extracted_cube = file['data'][rnd_t_HR:rnd_t_HR+int(t*self.scale), rnd_h:rnd_h+h, rnd_w:rnd_w+w].sum()
                        total_required = (self.scale * t * w * h)
                        fill_ratio = extracted_cube/total_required
                    vol_mask = None # mask not used for training, only for extracting LR and HR volume

                # get vol_mask during validation
                if self.opt['phase'] == 'val':
                    if self.mask_folder and self.opt['need_voxels'] and not self.opt['need_voxels']['tile_x_y']:
                        vol_mask = file['data'][rnd_t_HR:rnd_t_HR+int(t*self.scale), rnd_h:rnd_h+h, rnd_w:rnd_w+w]
                    else:
                        vol_mask = file['data'][()]

        # LR
        with h5py.File(os.path.join(self.in_folder, uid+'.h5'), 'r') as file:
            # print("LR data shape:", file['data'].shape) # same shape as mask 
            # extract volume cube during training, else return entire volume during validation
            if self.opt['phase'] == 'train':
                # print("LR z -> {}/{}:({}/{})+{}".format(rnd_t_HR, self.scale, rnd_t_HR, self.scale, t))
                # print("LR h -> {}:{}+{}".format(rnd_h, rnd_h, h))
                # print("LR w -> {}:{}+{}".format(rnd_w, rnd_w, w))
                vol_in = file['data'][round(rnd_t_HR/self.scale):round(rnd_t_HR/self.scale)+t,
                                      rnd_h:rnd_h+h, rnd_w:rnd_w+w]
            else:
                if self.mask_folder and self.opt['need_voxels'] and not self.opt['need_voxels']['tile_x_y']:
                    vol_in = file['data'][round(rnd_t_HR/self.scale):round(rnd_t_HR/self.scale)+t,
                                          rnd_h:rnd_h+h, rnd_w:rnd_w+w]
                else:
                    vol_in = file['data'][()]
        
        # HR (note: HR cube indexes will be same as mask cube indexes since the masks are for HR images)
        vol_tar = None
        if self.tar_folder:
            with h5py.File(os.path.join(self.tar_folder, uid+'.h5'), 'r') as file:
                # extract volume cube during training, else return entire volume during validation
                if self.opt['phase'] == 'train':
                    # print("HR z -> {}:{}+({}*{})".format(rnd_t_HR, rnd_t_HR, t, self.scale))
                    # print("HR h -> {}:{}+{}".format(rnd_h, rnd_h, h))
                    # print("HR w -> {}:{}+{}".format(rnd_w, rnd_w, w))
                    vol_tar = file['data'][rnd_t_HR:rnd_t_HR+int(t*self.scale), rnd_h:rnd_h+h, rnd_w:rnd_w+w]
                else:
                    if self.mask_folder and self.opt['need_voxels'] and not self.opt['need_voxels']['tile_x_y']:
                        vol_tar = file['data'][rnd_t_HR:rnd_t_HR+int(t*self.scale), rnd_h:rnd_h+h, rnd_w:rnd_w+w]
                    else:
                        vol_tar = file['data'][()]

        # creates an extra dimension. e.g. 32x64x64 becomes 1x32x64x64
        vol_in = np.expand_dims(vol_in, axis=0)
        vol_tar = np.expand_dims(vol_tar, axis=0)
        vol_mask = np.expand_dims(vol_mask, axis=0)
        # convert to tensors and also within a certain range of values as defined in 'self.ToTensor()' class
        vol_in, vol_tar = self.ToTensor(vol_in), self.ToTensor(vol_tar)
        vol_mask = self.ToTensor(vol_mask, raw_data_range = 1)

        # read LR spacings
        spacings = [] 
        if self.opt['phase'] == 'val':
            config_path = os.path.join(self.in_folder, uid + '.json')
            # we only need spacing information for the configuration file
            meta_data = util.read_config(config_path)
            spacings = meta_data['Spacing']

        # store information in dictionary 
        out_dict = {'LR': vol_in, 'HR': vol_tar, 'mask': vol_mask, 
                    'spacings': spacings, 'uid': uid
                    }
        return out_dict

    def __len__(self):
        return len(self.uids)



import os
import random
from sklearn.model_selection import train_test_split
import shutil
import nrrd


def write_txt_file(data_list, name, path):
    file_name = '{}.txt'.format(name)
    file_path = os.path.join(path, file_name)
    with open(file_path, 'w') as f:
        f.write('\n'.join(data_list))

def create_train_val_test(path_to_files, path_to_save_txt, external_test=None):
    all_files = os.listdir(path_to_files)
    if external_test:
        with open(external_test, 'r') as f:
            lines = f.readlines()
            test_cases = [l.rstrip() for l in lines]
    all_files = list(set(all_files)-set(test_cases))
    train_set, val_set = [], []
    nodule_cases, nonodule_cases = [], []
    for file in all_files:
        nod_type = file.split('_')[0]
        if nod_type == 'nod':
            nodule_cases.append(file)
        elif nod_type == 'nonod':
            nonodule_cases.append(file)
        else:
            raise ValueError('Unknown Type Found!')
    nodule_train, nodule_val = train_test_split(nodule_cases, test_size=0.15)
    nonodule_train, nonodule_val = train_test_split(nonodule_cases, test_size=0.15)
    train_set.extend(nodule_train)
    train_set.extend(nonodule_train)
    val_set.extend(nodule_val)
    val_set.extend(nonodule_val)
    random.shuffle(train_set)
    random.shuffle(val_set)
    write_txt_file(train_set, 'UCLA-Nas-train_nodules', path_to_save_txt)
    write_txt_file(val_set, 'UCLA-Nas-val_nodules', path_to_save_txt)

def merge_cases(path_to_cnd, path_to_save_merged):
    cnds = [cnd for cnd in os.listdir(path_to_cnd) if cnd != 'merged']
    for each_cnd in cnds:
        files = os.listdir(os.path.join(path_to_cnd, each_cnd))
        for each_file in files:
            src = os.path.join(path_to_cnd, each_cnd, each_file)
            dest = os.path.join(path_to_save_merged, '{}_cnd-{}.nrrd'.format(each_file.split('.nrrd')[0], each_cnd))
            shutil.copy(src, dest)
    print('All files copied!')

def generate_specific_uids(path_to_data):
    mapping_uids = {}
    for file in os.listdir(path_to_data):
        if file.endswith('.nrrd'):   
            condition = file.split('.nrrd')[0].split('_cnd-')[-1]
            if condition not in mapping_uids:
                mapping_uids[condition] = []
            if file not in mapping_uids[condition]:
                mapping_uids[condition].append(file)
    return mapping_uids

def generate_merged_train_test_uid(path_to_data, path_to_save_merged_txt, external_test):
    mapped_cnd = generate_specific_uids(path_to_data)
    train_set, test_set = [], []
    with open(external_test, 'r') as f:
        lines = f.readlines()
        test_cases = [l.rstrip() for l in lines]
    for cnd in mapped_cnd:
        all_files = mapped_cnd[cnd]
        test_cnd_cases = [file.split('.nrrd')[0]+'_cnd-{}.nrrd'.format(cnd) for file in test_cases]
        all_files = list(set(all_files)-set(test_cnd_cases))
        train_set.extend(all_files)
        test_set.extend(test_cnd_cases)
    print('Total train set samples:', len(train_set))
    print('Total test set samples:', len(test_set))
    random.shuffle(train_set)
    random.shuffle(test_set)
    write_txt_file(train_set, 'UCLA-Nas-merged_train_nodules', path_to_save_merged_txt)
    write_txt_file(test_set, 'UCLA-merged_test_nodules', path_to_save_merged_txt)

def check_size(vol_path, train_nods, path_to_save_merged_txt):
    updated_train = []
    files_skipped = 0
    with open(train_nods, 'r') as file:
        train_nodules = file.readlines()
        train_nodules = [line.strip() for line in train_nodules]
    for nod in train_nodules:
        img_path = os.path.join(vol_path, nod)
        data, header = nrrd.read(img_path)
        if data.shape[0] != 64 or data.shape[1] != 64 or data.shape[2] != 16:
            print('Skipping train file {} ...'.format(nod))
            files_skipped += 1
        else:
            updated_train.append(nod)
    random.shuffle(updated_train)
    print('{} files skipped'.format(files_skipped))
    write_txt_file(updated_train, 'UCLA-merged_train_nodules_v1', path_to_save_merged_txt)


if __name__ == '__main__':
    path_to_files = '/datasets/UCLA-Nas-NoduleVolume/unnormalized/k2_d100_st1.0'
    external_test = '/datasets/UCLA-Nas-NoduleVolume/uids/UCLA-test_uids.txt'
    path_to_save_txt = '/datasets/UCLA-Nas-NoduleVolume/uids'
    # create_train_val_test(path_to_files, path_to_save_txt, external_test)
    # =====================================================================
    path_to_cnd = '/datasets/UCLA-Nas-NoduleVolume/unnormalized'
    path_to_save_merged = '/datasets/UCLA-Nas-NoduleVolume/unnormalized/merged/data'
    # merge_cases(path_to_cnd, path_to_save_merged)
    # =====================================================================
    path_to_save_merged_txt = '/datasets/UCLA-Nas-NoduleVolume/unnormalized/merged'
    # generate_merged_train_test_uid(path_to_save_merged, path_to_save_merged_txt, external_test)
    # =====================================================================
    # check_size(path_to_save_merged, '/datasets/UCLA-Nas-NoduleVolume/unnormalized/merged/UCLA-Nas-merged_train_nodules.txt', path_to_save_merged_txt)



from radiomics import featureextractor
import numpy as np 
import tempfile
import nibabel as nib 
import SimpleITK as sitk
from scipy import ndimage
from skimage import measure
from lung_analysis_pipeline.utils import util 
import logging
radiomics_logger = logging.getLogger("radiomics")
radiomics_logger.setLevel(logging.ERROR)
glcm_logger = logging.getLogger("radiomics.glcm")
glcm_logger.setLevel(logging.ERROR)
logger = logging.getLogger('base')

# def extract_features(image, mask): 
#     # Create an instance of the PyRadiomics feature extractor
#     self.extractor = featureextractor.RadiomicsFeatureExtractor()

#     # Get the default params 
#     params = extractor.getParams()

#     # Modify the parameters as desired
#     params['normalize'] = True
#     params['resampledPixelSpacing'] = [1.0, 1.0, 1.0]
#     params['binWidth'] = 25

#     # Extract features using the image and mask
#     features = extractor.execute(image, mask)
#     logger.info('Radiomic feature extractor is created.')

class radiomic_feature_extractor(): 
    def __init__(self, opt): 
        self.opt = opt 
        self.extractor = featureextractor.RadiomicsFeatureExtractor()
        # First define the settings
        settings = self.extractor.settings
        settings['binWidth'] = 20
        settings['sigma'] = [1, 2, 3]

        # Instantiate the extractor
        extractor = featureextractor.RadiomicsFeatureExtractor(**settings)  # ** 'unpacks' the dictionary in the function call
        logger.info('Radiomic feature extractor is created.')
        # logger.info('Extraction parameters:\n\t', extractor.settings)
        # logger.info('Enabled filters:\n\t', extractor.enabledImagetypes)  # Still the default parameters
        # logger.info('Enabled features:\n\t', extractor.enabledFeatures)  # Still the default parameters
        

    def run_test(self, input_data, roi, threshold_size=0): 
        if roi == 'segmentation' or roi == 'user': 
            return self.extract_from_entire_mask(input_data, roi, threshold_size=threshold_size)
        elif roi == 'detection': 
            return self.extract_from_bbox(input_data, threshold_size=threshold_size)
        else: 
            raise ValueError('Feature extraction ROI source not supported. Need to specify user/detection/segmentation.')
    
    def extract_from_bbox(self, input_data, threshold_size=0):
        image = input_data['volume'][0][0].cpu().numpy().astype(np.float32)
        uid = input_data['uid'][0]

        # input_data['detection_bboxes'] = torch.from_numpy(detections)[None] 
        detection_bboxes = input_data['detection_bboxes'][0].cpu().numpy()

        features = [] 

        # Iterate through the detection results 
        for i, bbox in enumerate(detection_bboxes): 
            bbox = bbox.astype(int)
            _, z, y, x, d, _ = bbox 
            # Calculate starting and ending indices for each dimension
            start_z = max(0, z - d // 2)
            end_z = min(image.shape[0], z + d // 2 + 1)
            start_y = max(0, y - d // 2)
            end_y = min(image.shape[1], y + d // 2 + 1)
            start_x = max(0, x - d // 2)
            end_x = min(image.shape[2], x + d // 2 + 1)

            roi = image[start_z:end_z, start_y:end_y, start_x:end_x]
            roi_mask = np.ones_like(roi)

            # Extract features from the ROI
            feature = self.extractor.execute(sitk.GetImageFromArray(roi), sitk.GetImageFromArray(roi_mask))
            feature['NoduleFeatureID'] = f"{uid}-{i-1}"
            feature.move_to_end('NoduleFeatureID', last=False)

            features.append(util.convert_numpy_to_list(feature))
        
        input_data['radiomics'] = features
        return input_data

    def extract_from_entire_mask(self, input_data, roi_source, threshold_size=0): 
        image = input_data['volume'][0][0].cpu().numpy().astype(np.float32)
        uid = input_data['uid'][0]
        if roi_source == 'segmentation':
            mask = input_data['seg_mask'][0].cpu().numpy().astype(np.uint8)
        elif roi_source == 'user':
            mask = input_data['gt_mask'][0].cpu().numpy().astype(np.uint8)
        # Extract connected compontents in mask 
        labeled_mask, num_components = measure.label(mask, return_num=True)

        features = [] # list of OrderedDict 
        
        for i in range(1, num_components + 1):
            # Extract the binary mask for the current component
            component_mask = (labeled_mask == i)
            # Check if the size of the current component is larger than the threshold
            if np.sum(component_mask) < threshold_size:
                continue
            # Calculate the centroid of the current component
            # z, y, x = np.round(np.array(ndimage.center_of_mass(component_mask))).astype(int)
            # centroid = np.array([z, y, x])
            
            # Get a bounding box of this connected component and extract roi and mask 
            z, y, x = np.where(component_mask)
            start_z, end_z = np.min(z), np.max(z) + 1
            start_y, end_y = np.min(y), np.max(y) + 1
            start_x, end_x = np.min(x), np.max(x) + 1

            roi = image[start_z:end_z, start_y:end_y, start_x:end_x]
            roi_mask = mask[start_z:end_z, start_y:end_y, start_x:end_x]

            # Extract features from the ROI
            feature = self.extractor.execute(sitk.GetImageFromArray(roi), sitk.GetImageFromArray(roi_mask))
            feature['NoduleFeatureID'] = f"{uid}-{i-1}"
            feature.move_to_end('NoduleFeatureID', last=False)

            features.append(util.convert_numpy_to_list(feature))
        
        input_data['radiomics'] = features
        return input_data
    
    





import h5py
import SimpleITK as sitk
import os
import json
import nrrd
import numpy as np


def read_config(config_path):
    config_path = os.path.join(config_path) 
    with open(config_path, 'r') as f:
        config = json.load(f)
    meta_data = {}
    meta_data['Spacing'] = [float(i) for i in config['Spacing'].split()]
    meta_data['Orientation'] = [float(i) for i in config['Orientation'].split()] 
    meta_data['Origin'] = [float(i) for i in config['Origin'].split()]
    header = {'units': ['mm', 'mm', 'mm'], 'spacings': meta_data['Spacing']} 
    return header

"""
Histogram matching following the method developed on
Nyul et al 2001 (ITK implementation)
inputs:
- mov_scan: np.array containing the image to normalize
- ref_scan np.array containing the reference image
- histogram levels
- number of matched points
- Threshold Mean setting
outputs:
- histogram matched image
"""
def histogram_matching(lr_scan_path, ref_scan_path, UID, save_path, histogram_levels=2048, match_points=100, set_th_mean=True, split_uid_for_ref=True):
    with open(UID, 'r') as f:
        lines = f.readlines()
        uids = [l.rstrip() for l in lines]
    if not os.path.exists(save_path):
        os.makedirs(save_path)

    # read each uid
    for uid in uids:
        if split_uid_for_ref:
            uid_for_ref = uid.split('_')[0]
        else:
            uid_for_ref = uid
        print('Reading UID:', uid)
        LR_path = os.path.join(lr_scan_path, '{}.h5'.format(uid))
        HR_path = os.path.join(ref_scan_path, '{}.h5'.format(uid_for_ref))
        with h5py.File(LR_path, 'r') as file:
            mov_scan = file['data'][()]
        with h5py.File(HR_path, 'r') as file:
            ref_scan = file['data'][()]

        header = os.path.join(ref_scan_path, '{}.json'.format(uid_for_ref))
        volume_save_path = os.path.join(save_path, '{}.nrrd'.format(uid))
        spacing_info = read_config(header)
 
        # convert np arrays into itk image objects
        ref = sitk.GetImageFromArray(ref_scan.astype('float32'))
        mov = sitk.GetImageFromArray(mov_scan.astype('float32'))
        # perform histogram matching
        caster = sitk.CastImageFilter()
        caster.SetOutputPixelType(ref.GetPixelID())
        matcher = sitk.HistogramMatchingImageFilter()
        matcher.SetNumberOfHistogramLevels(histogram_levels)
        matcher.SetNumberOfMatchPoints(match_points)
        matcher.SetThresholdAtMeanIntensity(set_th_mean)
        matched_vol = matcher.Execute(mov, ref)
        nda = sitk.GetArrayFromImage(matched_vol)
        # clip to desired range
        nda = nda.astype(np.int16)
        nda_clipped = np.clip(nda, 0, 1500)
        nda_clipped = nda_clipped.astype(np.int16)
        # nrrd.write(volume_save_path, nda, spacing_info, index_order='C')
        nrrd.write(volume_save_path, nda_clipped, spacing_info, index_order='C')
    print('HM DONE!')


if __name__ == '__main__':
    dataroot_LR = '/datasets/data_st1.0_merged' # '/dingo_data_leihaowei/nastaran_test_h5/k2_d10_st1.0'
    dataroot_HR = '/datasets/reference/k2_d100_st1' # '/dingo_data_leihaowei/nastaran_test_h5/k2_d100_st1.0'
    UID = '/datasets/uids/70-30_split/ucla_testUID_70_30_split.txt' # '/datasets/nodule_annotations/ucla_nastaran_subset_32_test_uid.txt'
    save_path = '../../results/test_HM-70-30-Split_tileStitch-32x64x64_Ref-k2d100/merged_cases/test_st1.0_norm' # train_CNN-SRResNet-EQMergedData_tileStitch-32x64x64_Ref-k2d100
    histogram_matching(dataroot_LR, dataroot_HR, UID, save_path)







from __future__ import absolute_import
import sys
sys.path.append('..')
sys.path.append('.')
import torch
import torch.nn as nn
import torch.nn.init as init
from torch.autograd import Variable
import numpy as np
from pdb import set_trace as st
from skimage import color
from IPython import embed
from . import pretrained_networks as pn
from ..util import util
# from util import util

# Off-the-shelf deep network
class PNet(nn.Module):
    '''Pre-trained network with all channels equally weighted by default'''
    def __init__(self, pnet_type='vgg', pnet_rand=False, use_gpu=True):
        super(PNet, self).__init__()

        self.use_gpu = use_gpu
        self.pnet_type = pnet_type
        self.pnet_rand = pnet_rand
        self.shift = torch.autograd.Variable(torch.Tensor([-.030, -.088, -.188]).view(1,3,1,1))
        self.scale = torch.autograd.Variable(torch.Tensor([.458, .448, .450]).view(1,3,1,1))
        
        if(self.pnet_type in ['vgg','vgg16']):
            self.net = pn.vgg16(pretrained=not self.pnet_rand,requires_grad=False)
        elif(self.pnet_type=='alex'):
            self.net = pn.alexnet(pretrained=not self.pnet_rand,requires_grad=False)
        elif(self.pnet_type[:-2]=='resnet'):
            self.net = pn.resnet(pretrained=not self.pnet_rand,requires_grad=False, num=int(self.pnet_type[-2:]))
        elif(self.pnet_type=='squeeze'):
            self.net = pn.squeezenet(pretrained=not self.pnet_rand,requires_grad=False)

        self.L = self.net.N_slices

        if(use_gpu):
            self.net.cuda()
            self.shift = self.shift.cuda()
            self.scale = self.scale.cuda()

    def forward(self, in0, in1, retPerLayer=False):
        in0_sc = (in0 - self.shift.expand_as(in0))/self.scale.expand_as(in0)
        in1_sc = (in1 - self.shift.expand_as(in0))/self.scale.expand_as(in0)

        outs0 = self.net.forward(in0_sc)
        outs1 = self.net.forward(in1_sc)

        if(retPerLayer):
            all_scores = []
        for (kk,out0) in enumerate(outs0):
            cur_score = (1.-util.cos_sim(outs0[kk],outs1[kk]))
            if(kk==0):
                val = 1.*cur_score
            else:
                # val = val + self.lambda_feat_layers[kk]*cur_score
                val = val + cur_score
            if(retPerLayer):
                all_scores+=[cur_score]

        if(retPerLayer):
            return (val, all_scores)
        else:
            return val

"""
Learned perceptual metric
"""
class PNetLin(nn.Module):
    def __init__(self, pnet_type='vgg', pnet_rand=False, pnet_tune=False, use_dropout=True, use_gpu=True, spatial=False, version='0.1'):
        super(PNetLin, self).__init__()
        print("PNetLin model instantiating")
        print("-"*40)
        self.use_gpu = use_gpu
        self.pnet_type = pnet_type
        self.pnet_tune = pnet_tune
        self.pnet_rand = pnet_rand
        self.spatial = spatial
        self.version = version

        if(self.pnet_type in ['vgg','vgg16']):
            net_type = pn.vgg16 # get vgg16 architecture class, divided into chunk of slices as named tuple
            """
            # to check architecture
            net = net_type()
            print("chuncked vgg16:", net)
            """
            self.chns = [64,128,256,512,512]
        elif(self.pnet_type=='alex'):
            net_type = pn.alexnet
            self.chns = [64,192,384,256,256]
        elif(self.pnet_type=='squeeze'):
            net_type = pn.squeezenet
            self.chns = [64,128,256,384,384,512,512]

        if(self.pnet_tune):
            print("vgg16 will be tuned (requires_grad=True) & will be part of PNetLin model architecture")
            self.net = net_type(pretrained=not self.pnet_rand,requires_grad=True)
        else:
            print("vgg16 will not be tuned (requires_grad=False) & will not be part of PNetLin model architecture")
            # 'not self.pnet_rand' = not False = True, so we download pretrained vgg26
            self.net = [net_type(pretrained=not self.pnet_rand,requires_grad=False),]

        # print("self.net:", self.net)
        # create linear layers, mapping to one channel
        self.lin0 = NetLinLayer(self.chns[0],use_dropout=use_dropout) # (64, 1)
        self.lin1 = NetLinLayer(self.chns[1],use_dropout=use_dropout) # (128, 1)
        self.lin2 = NetLinLayer(self.chns[2],use_dropout=use_dropout) # (256, 1)
        self.lin3 = NetLinLayer(self.chns[3],use_dropout=use_dropout) # (512, 1)
        self.lin4 = NetLinLayer(self.chns[4],use_dropout=use_dropout) # (512, 1)
        # create a list of linear layers
        self.lins = [self.lin0,self.lin1,self.lin2,self.lin3,self.lin4]
        
        # if 'pnet_type=seqeeze', we need 7 layers so we add two extra
        if(self.pnet_type=='squeeze'):
            self.lin5 = NetLinLayer(self.chns[5],use_dropout=use_dropout)
            self.lin6 = NetLinLayer(self.chns[6],use_dropout=use_dropout)
            self.lins += [self.lin5,self.lin6]

        # what is this shift and scale ?
        self.shift = torch.autograd.Variable(torch.Tensor([-.030, -.088, -.188]).view(1,3,1,1))
        self.scale = torch.autograd.Variable(torch.Tensor([.458, .448, .450]).view(1,3,1,1))

        if(use_gpu):
            # if model is being tuned, put it on cuda
            if(self.pnet_tune):
                self.net.cuda()
            else:
                # put the first element of list (self.net), which is the chunked vgg16, on cuda
                self.net[0].cuda()
              
            # put shift & scale autograd variable on cuda
            self.shift = self.shift.cuda()
            self.scale = self.scale.cuda()
            # put the linear layers on cuda
            self.lin0.cuda()
            self.lin1.cuda()
            self.lin2.cuda()
            self.lin3.cuda()
            self.lin4.cuda()
            # if 'pnet_type' is 'squeeze', put the additional two layers on cuda
            if(self.pnet_type=='squeeze'):
                self.lin5.cuda()
                self.lin6.cuda()

    def forward(self, in0, in1):
        in0_sc = (in0 - self.shift.expand_as(in0))/self.scale.expand_as(in0)
        in1_sc = (in1 - self.shift.expand_as(in0))/self.scale.expand_as(in0)

        if(self.version=='0.0'):
            # v0.0 - original release had a bug, where input was not scaled
            in0_input = in0
            in1_input = in1
        else:
            # v0.1
            in0_input = in0_sc
            in1_input = in1_sc

        if(self.pnet_tune):
            outs0 = self.net.forward(in0_input)
            outs1 = self.net.forward(in1_input)
        else:
            outs0 = self.net[0].forward(in0_input)
            outs1 = self.net[0].forward(in1_input)

        feats0 = {}
        feats1 = {}
        diffs = [0]*len(outs0)

        for (kk,out0) in enumerate(outs0):
            feats0[kk] = util.normalize_tensor(outs0[kk])
            feats1[kk] = util.normalize_tensor(outs1[kk])
            diffs[kk] = (feats0[kk]-feats1[kk])**2

        if self.spatial:
            lin_models = [self.lin0, self.lin1, self.lin2, self.lin3, self.lin4]
            if(self.pnet_type=='squeeze'):
                lin_models.extend([self.lin5, self.lin6])
            res = [lin_models[kk].model(diffs[kk]) for kk in range(len(diffs))]
            return res
			
        val = torch.mean(torch.mean(self.lin0.model(diffs[0]),dim=3),dim=2)
        val = val + torch.mean(torch.mean(self.lin1.model(diffs[1]),dim=3),dim=2)
        val = val + torch.mean(torch.mean(self.lin2.model(diffs[2]),dim=3),dim=2)
        val = val + torch.mean(torch.mean(self.lin3.model(diffs[3]),dim=3),dim=2)
        val = val + torch.mean(torch.mean(self.lin4.model(diffs[4]),dim=3),dim=2)
        if(self.pnet_type=='squeeze'):
            val = val + torch.mean(torch.mean(self.lin5.model(diffs[5]),dim=3),dim=2)
            val = val + torch.mean(torch.mean(self.lin6.model(diffs[6]),dim=3),dim=2)

        val = val.view(val.size()[0],val.size()[1],1,1)
        return val

class Dist2LogitLayer(nn.Module):
    ''' takes 2 distances, puts through fc layers, spits out value between [0,1] (if use_sigmoid is True) '''
    def __init__(self, chn_mid=32,use_sigmoid=True):
        super(Dist2LogitLayer, self).__init__()
        layers = [nn.Conv2d(5, chn_mid, 1, stride=1, padding=0, bias=True),]
        layers += [nn.LeakyReLU(0.2,True),]
        layers += [nn.Conv2d(chn_mid, chn_mid, 1, stride=1, padding=0, bias=True),]
        layers += [nn.LeakyReLU(0.2,True),]
        layers += [nn.Conv2d(chn_mid, 1, 1, stride=1, padding=0, bias=True),]
        if(use_sigmoid):
            layers += [nn.Sigmoid(),]
        self.model = nn.Sequential(*layers)

    def forward(self,d0,d1,eps=0.1):
        return self.model.forward(torch.cat((d0,d1,d0-d1,d0/(d1+eps),d1/(d0+eps)),dim=1))

class BCERankingLoss(nn.Module):
    def __init__(self, use_gpu=True, chn_mid=32):
        super(BCERankingLoss, self).__init__()
        self.use_gpu = use_gpu
        self.net = Dist2LogitLayer(chn_mid=chn_mid)
        self.parameters = list(self.net.parameters())
        self.loss = torch.nn.BCELoss()
        self.model = nn.Sequential(*[self.net])

        if(self.use_gpu):
            self.net.cuda()

    def forward(self, d0, d1, judge):
        per = (judge+1.)/2.
        if(self.use_gpu):
            per = per.cuda()
        self.logit = self.net.forward(d0,d1)
        return self.loss(self.logit, per)

"""
A single linear layer, w/wo dropout which does a 1x1 conv
"""
class NetLinLayer(nn.Module):
    def __init__(self, chn_in, chn_out=1, use_dropout=False):
        super(NetLinLayer, self).__init__()
        layers = [nn.Dropout(),] if(use_dropout) else []
        layers += [nn.Conv2d(chn_in, chn_out, 1, stride=1, padding=0, bias=False),]
        self.model = nn.Sequential(*layers)


# L2, DSSIM metrics
class FakeNet(nn.Module):
    def __init__(self, use_gpu=True, colorspace='Lab'):
        super(FakeNet, self).__init__()
        self.use_gpu = use_gpu
        self.colorspace=colorspace

class L2(FakeNet):

    def forward(self, in0, in1):
        assert(in0.size()[0]==1) # currently only supports batchSize 1

        if(self.colorspace=='RGB'):
            (N,C,X,Y) = in0.size()
            value = torch.mean(torch.mean(torch.mean((in0-in1)**2,dim=1).view(N,1,X,Y),dim=2).view(N,1,1,Y),dim=3).view(N)
            return value
        elif(self.colorspace=='Lab'):
            value = util.l2(util.tensor2np(util.tensor2tensorlab(in0.data,to_norm=False)), 
                util.tensor2np(util.tensor2tensorlab(in1.data,to_norm=False)), range=100.).astype('float')
            ret_var = Variable( torch.Tensor((value,) ) )
            if(self.use_gpu):
                ret_var = ret_var.cuda()
            return ret_var

class DSSIM(FakeNet):

    def forward(self, in0, in1):
        assert(in0.size()[0]==1) # currently only supports batchSize 1

        if(self.colorspace=='RGB'):
            value = util.dssim(1.*util.tensor2im(in0.data), 1.*util.tensor2im(in1.data), range=255.).astype('float')
        elif(self.colorspace=='Lab'):
            value = util.dssim(util.tensor2np(util.tensor2tensorlab(in0.data,to_norm=False)), 
                util.tensor2np(util.tensor2tensorlab(in1.data,to_norm=False)), range=100.).astype('float')
        ret_var = Variable( torch.Tensor((value,) ) )
        if(self.use_gpu):
            ret_var = ret_var.cuda()
        return ret_var

"""
Print the network architecture and number of parameters for that model
"""
def print_network(net):
    num_params = 0
    for param in net.parameters():
        num_params += param.numel()
    print('Network',net)
    print('Total number of parameters: %d' % num_params)



try:
    from lung_analysis_pipeline.NoduleDetect_SANet.utils.pybox import *
except ImportError:
    print('Warning: C++ module import failed! This should only happen in deployment')
    from lung_analysis_pipeline.NoduleDetect_SANet.utils.util import py_nms as torch_nms
    from lung_analysis_pipeline.NoduleDetect_SANet.utils.util import py_box_overlap as torch_overlap
import random
from .rpn_nms import rpn_encode
from torch.autograd import Variable


def make_one_rpn_target(cfg, mode, input, window, truth_box, truth_label):
    """
    Generate region proposal targets for one batch

    cfg: dict, for hyper-parameters
    mode: string, which phase/mode is used currently
    input: 5D torch tensor of [batch, channel, z, y, x], original input to the network
    window: list of anchor bounding boxes, [z, y, x, d, h, w]
    truth_box: list of ground truth bounding boxes, [z, y, x, d, h, w]
    truth_label: list of grount truth class label for each object in the correponding truth_box

    return torch tensors
    label: positive or negative (1 or 0) for each anchor box
    label_assign: index of the ground truth box, to which the anchor box is matched to
    label_weight: class weight for each sample, zero means current sample is protected,
                  and won't contribute to loss
    target: bounding box regression terms
    target_weight: weight for each regression term, by default it should all be ones
    """

    num_neg = cfg['num_neg']
    num_window = len(window)
    label = np.zeros((num_window, ), np.float32)
    label_assign = np.zeros((num_window, ), np.int32) - 1
    label_weight = np.zeros((num_window, ), np.float32)
    target = np.zeros((num_window, 6), np.float32)
    target_weight = np.zeros((num_window, ), np.float32)


    num_truth_box = len(truth_box)
    if num_truth_box:

        _, depth, height, width = input.size()

        # Get sure background anchor boxes
        overlap = torch_overlap(window, truth_box)

        # For each anchor box, get the index of the ground truth box that
        # has the largest IoU with it
        argmax_overlap = np.argmax(overlap, 1)

        # For each anchor box, get the IoU of the ground truth box that
        # has the largest IoU with it
        max_overlap = overlap[np.arange(num_window), argmax_overlap]

        # The anchor box is a sure background, if its largest IoU is less than
        # a threshold
        bg_index = np.where(max_overlap < cfg['rpn_train_bg_thresh_high'])[0]
        label[bg_index] = 0
        label_weight[bg_index] = 1

        # The anchor box is a sure foreground, if its largest IoU is larger or 
        # equal than a threshold
        fg_index = np.where(max_overlap >= cfg['rpn_train_fg_thresh_low'])[0]
        label[fg_index] = 1
        label_weight[fg_index] = 1
        label_assign[...] = argmax_overlap

        # In case no anchor box that overlaps with the ground truth box meets the threshold, 
        # for each ground truth box, we include anchor box that has the highest IoU with it, 
        # include multiple maxs if there exists more than one anchor box
        argmax_overlap = np.argmax(overlap,0)
        max_overlap = overlap[argmax_overlap,np.arange(num_truth_box)]
        argmax_overlap, a = np.where(overlap == max_overlap)

        fg_index = argmax_overlap

        label[fg_index] = 1
        label_weight[fg_index] = 1
        label_assign[fg_index] = a

        # In case one ground truth box within one batch has way too many positive anchors, 
        # which may affect the sample in the loss fucntion,
        # we only random one positive anchor for each ground truth box
        fg_index = np.where(label != 0)[0]
        idx = random.sample(range(len(fg_index)), 1)
        label[fg_index] = 0
        label_weight[fg_index] = 0
        fg_index = fg_index[idx]
        label[fg_index] = 1
        label_weight[fg_index] = 1


        # Prepare regression terms for each positive anchor
        fg_index = np.where(label != 0)[0]
        target_window = window[fg_index]
        target_truth_box = truth_box[label_assign[fg_index]]
        target[fg_index] = rpn_encode(target_window, target_truth_box, cfg['box_reg_weight'])
        target_weight[fg_index] = 1

        # This should by no means be used, left just in case
        invalid_truth_label = np.where(truth_label < 0)[0]
        invalid_index = np.isin(label_assign, invalid_truth_label) & (label!=0)
        label_weight[invalid_index] = 0
        target_weight[invalid_index] = 0

        if mode in ['train']:
            fg_index = np.where( (label_weight!=0) & (label!=0))[0]
            bg_index = np.where( (label_weight!=0) & (label==0))[0]

            # Random sample num_neg negative anchor boxes first
            # This is very strange, but it works well in practice
            # It makes the use of hard negative example mining loss, not 
            # actually hard negative example mining.
            label_weight[bg_index] = 0
            idx = random.sample(range(len(bg_index)), min(num_neg, len(bg_index)))
            bg_index = bg_index[idx]

            # Calculate weight for class balance
            num_fg = max(1, len(fg_index))
            num_bg = len(bg_index)
            label_weight[bg_index] = float(num_fg)/num_bg

        target_weight[fg_index] = label_weight[fg_index]
    else:
        # if there is no ground truth box in this batch
        label_weight[...] = 1

        if mode in ['train']:
            bg_index = np.where((label_weight!=0) & (label==0))[0]

            label_weight[bg_index] = 0
            idx = random.sample(range(len(bg_index)), min(num_neg, len(bg_index)))
            bg_index = bg_index[idx]
            label_weight[bg_index] = 1.0 / len(bg_index)


    label = Variable(torch.from_numpy(label)).cuda()
    label_assign = Variable(torch.from_numpy(label_assign)).cuda()
    label_weight = Variable(torch.from_numpy(label_weight)).cuda()
    target = Variable(torch.from_numpy(target)).cuda()
    target_weight = Variable(torch.from_numpy(target_weight)).cuda()
    return  label, label_assign, label_weight, target, target_weight


def make_rpn_target(cfg, mode, inputs, window, truth_boxes, truth_labels):
    rpn_labels = []
    rpn_label_assigns = []
    rpn_label_weights = []
    rpn_targets = []
    rpn_targets_weights = []

    batch_size = len(inputs)
    for b in range(batch_size):
        input = inputs[b]
        truth_box   = truth_boxes[b]
        truth_label = truth_labels[b]

        rpn_label, rpn_label_assign, rpn_label_weight, rpn_target, rpn_targets_weight = \
            make_one_rpn_target(cfg, mode, input, window, truth_box, truth_label)

        rpn_labels.append(rpn_label.view(1, -1))
        rpn_label_assigns.append(rpn_label_assign.view(1, -1))
        rpn_label_weights.append(rpn_label_weight.view(1, -1))
        rpn_targets.append(rpn_target.view(1, -1, 6))
        rpn_targets_weights.append(rpn_targets_weight.view(1, -1))


    rpn_labels          = torch.cat(rpn_labels, 0)
    rpn_label_assigns   = torch.cat(rpn_label_assigns, 0)
    rpn_label_weights   = torch.cat(rpn_label_weights, 0)
    rpn_targets         = torch.cat(rpn_targets, 0)
    rpn_targets_weights = torch.cat(rpn_targets_weights, 0)

    return rpn_labels, rpn_label_assigns, rpn_label_weights, rpn_targets, rpn_targets_weights



import math
from collections import Counter
from collections import defaultdict
import torch
from torch.optim.lr_scheduler import _LRScheduler


class MultiStepLR_Restart(_LRScheduler):
    def __init__(self, optimizer, milestones, restarts=None, weights=None, gamma=0.1,
                 clear_state=False, last_epoch=-1):
        self.milestones = Counter(milestones)
        self.gamma = gamma
        self.clear_state = clear_state
        self.restarts = restarts if restarts else [0]
        self.restart_weights = weights if weights else [1]
        assert len(self.restarts) == len(
            self.restart_weights), 'restarts and their weights do not match.'
        super(MultiStepLR_Restart, self).__init__(optimizer, last_epoch)

    def get_lr(self):
        if self.last_epoch in self.restarts:
            if self.clear_state:
                self.optimizer.state = defaultdict(dict)
            weight = self.restart_weights[self.restarts.index(self.last_epoch)]
            return [group['initial_lr'] * weight for group in self.optimizer.param_groups]
        if self.last_epoch not in self.milestones:
            return [group['lr'] for group in self.optimizer.param_groups]
        return [
            group['lr'] * self.gamma**self.milestones[self.last_epoch]
            for group in self.optimizer.param_groups
        ]


class CosineAnnealingLR_Restart(_LRScheduler):
    def __init__(self, optimizer, T_period, restarts=None, weights=None, eta_min=0, last_epoch=-1):
        self.T_period = [int(t) for t in T_period]
        self.T_max = self.T_period[0]  # current T period
        self.eta_min = eta_min
        self.restarts = restarts if restarts else [0]
        self.restart_weights = weights if weights else [1]
        self.last_restart = 0
        assert len(self.restarts) == len(
            self.restart_weights), 'restarts and their weights do not match.'
        super(CosineAnnealingLR_Restart, self).__init__(optimizer, last_epoch)

    def get_lr(self):
        if self.last_epoch == 0:
            return self.base_lrs
        elif self.last_epoch in self.restarts:
            self.last_restart = self.last_epoch
            self.T_max = self.T_period[self.restarts.index(self.last_epoch) + 1]
            weight = self.restart_weights[self.restarts.index(self.last_epoch)]
            return [group['initial_lr'] * weight for group in self.optimizer.param_groups]
        elif (self.last_epoch - self.last_restart - 1 - self.T_max) % (2 * self.T_max) == 0:
            return [
                group['lr'] + (base_lr - self.eta_min) * (1 - math.cos(math.pi / self.T_max)) / 2
                for base_lr, group in zip(self.base_lrs, self.optimizer.param_groups)
            ]
        return [(1 + math.cos(math.pi * (self.last_epoch - self.last_restart) / self.T_max)) /
                (1 + math.cos(math.pi * ((self.last_epoch - self.last_restart) - 1) / self.T_max)) *
                (group['lr'] - self.eta_min) + self.eta_min
                for group in self.optimizer.param_groups]


if __name__ == "__main__":
    optimizer = torch.optim.Adam([torch.zeros(3, 64, 3, 3)], lr=2e-4, weight_decay=0,
                                 betas=(0.9, 0.99))
    ##############################
    # MultiStepLR_Restart
    ##############################
    ## Original
    lr_steps = [20e3, 40e3, 60e3, 80e3]
    restarts = None
    restart_weights = None

    # ## two

    # lr_steps = [20e3, 40e3, 80e3, 120e3, 160e3, 180e3, 200e3, 240e3, 280e3]
    # restarts = [160e3]
    # restart_weights = [1]
    #
    # ## four
    # lr_steps = [10e3, 20e3, 40e3, 60e3, 80e3, 90e3, 100e3, 120e3, 140e3, \
    #             170e3, 180e3, 200e3, 220e3, 240e3, 250e3, 260e3, 280e3, 300e3]
    # restarts = [80e3, 160e3, 240e3]
    # restart_weights = [1, 1, 1]

    scheduler = MultiStepLR_Restart(optimizer, lr_steps, restarts, restart_weights, gamma=0.5,
                                    clear_state=False)

    ##############################
    # Cosine Annealing Restart
    ##############################

    ## one
    # T_period = [160e3]
    #
    # ## four
    # T_period = [20e3, 20e3, 40e3, 80e3]
    # restarts = [20e3, 40e3, 80e3]
    # restart_weights = [1, 1, 1]

    # three
    # T_period = [1e3, 1e3, 1e3, 1e3, 2e3, 2e3, 2e3]
    # restarts = [1e3, 2e3, 3e3, 4e3, 6e3, 8e3]
    # restart_weights = [1, 1, 1, 1, 1, 1]
    # eta_min = 2e-6
    # scheduler = CosineAnnealingLR_Restart(optimizer, T_period, eta_min=eta_min, restarts=restarts,
    #                                       weights=restart_weights)

    ##############################
    # Draw figure
    ##############################
    N_iter = int(100e3)
    lr_l = list(range(N_iter))
    for i in range(N_iter):
        scheduler.step()
        current_lr = optimizer.param_groups[0]['lr']
        lr_l[i] = current_lr

    import matplotlib as mpl
    from matplotlib import pyplot as plt
    import matplotlib.ticker as mtick
    mpl.style.use('default')
    import seaborn
    seaborn.set(style='whitegrid')
    seaborn.set_context('paper')

    plt.figure(1)
    plt.subplot(111)
    plt.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))
    plt.title('Title', fontsize=16, color='k')
    plt.plot(list(range(N_iter)), lr_l, linewidth=1.5, label='learning rate scheme')
    legend = plt.legend(loc='upper right', shadow=False)
    ax = plt.gca()
    labels = ax.get_xticks().tolist()
    for k, v in enumerate(labels):
        labels[k] = str(int(v / 1000)) + 'K'
    ax.set_xticklabels(labels)
    ax.yaxis.set_major_formatter(mtick.FormatStrFormatter('%.1e'))
    # plt.yscale('log')

    ax.set_ylabel('Learning rate')
    ax.set_xlabel('Iteration')
    fig = plt.gcf()
    plt.show()


'''
Configs for training & testing
Written by Whalechen
'''

import argparse

def parse_opts():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--data_root',
        default='/workspace/NoduleSeg_MedicalNet/data',
        type=str,
        help='Root directory path of data')
    parser.add_argument(
        '--img_train_list',
        default='/data/train_lidc_nodule_patchonly_ft.txt',
        type=str,
        help='Path for image list file')
    parser.add_argument(
        '--img_val_list',
        default='/data/val_lidc_nodule_patchonly_ft.txt',
        type=str,
        help='Path for image list file')
    parser.add_argument(
        '--img_test_list',
        default='data/test_lidc_subset.txt',
        type=str,
        help='Path for image list file')
    parser.add_argument(
        '--n_seg_classes',
        default=1,
        type=int,
        help="Number of segmentation classes"
    )
    parser.add_argument(
        '--learning_rate',  # set to 0.001 when finetune
        default=0.001,
        type=float,
        help=
        'Initial learning rate (divided by 10 while training by lr scheduler)')
    parser.add_argument(
        '--num_workers',
        default=16,
        type=int,
        help='Number of jobs')
    parser.add_argument(
        '--batch_size', default=16, type=int, help='Batch Size')
    parser.add_argument(
        '--phase', default='train', type=str, help='Phase of train or test')
    parser.add_argument(
        '--save_intervals',
        default=10,
        type=int,
        help='Interation for saving model')
    parser.add_argument(
        '--n_epochs',
        default=200,
        type=int,
        help='Number of total epochs to run')
    parser.add_argument(
        '--input_D',
    default=32,
        type=int,
        help='Input size of depth')
    parser.add_argument(
        '--input_H',
        default=64,
        type=int,
        help='Input size of height')
    parser.add_argument(
        '--input_W',
        default=64,
        type=int,
        help='Input size of width')
    parser.add_argument(
        '--resume_path',
        default='',
        type=str,
        help=
        'Path for resume model.'
    )
    parser.add_argument(
        '--pretrain_path',
        default='',
        type=str,
        help=
        'Path for pretrained model.'
    )
    parser.add_argument(
        '--new_layer_names',
        default=[], # 'conv_segment' 
        type=list,
        help='New layer except for backbone')
    parser.add_argument(
        '--no_cuda', action='store_true', help='If true, cuda is not used.')
    parser.set_defaults(no_cuda=False)
    parser.add_argument(
        '--gpu_id',
        nargs='+',
        type=int,              
        help='Gpu id lists')
    parser.add_argument(
        '--model',
        default='resnet',
        type=str,
        help='(resnet | preresnet | wideresnet | resnext | densenet | ')
    parser.add_argument(
        '--model_depth',
        default=34,
        type=int,
        help='Depth of resnet (10 | 18 | 34 | 50 | 101)')
    parser.add_argument(
        '--resnet_shortcut',
        default='B',
        type=str,
        help='Shortcut type of resnet (A | B)')
    parser.add_argument(
        '--manual_seed', default=1, type=int, help='Manually set random seed')
    parser.add_argument(
        '--ci_test', action='store_true', help='If true, ci testing is used.')
    args = parser.parse_args()
    args.save_folder = "./trails_UNet-FullDataset/models/{}_{}".format(args.model, args.model_depth)
    
    return args



import os
import scipy.ndimage as ndimage
import matplotlib.pyplot as plt
from skimage import measure, morphology, segmentation
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
import nrrd
import cv2
import h5py
import numpy as np


def generate_markers(image):
    image = (image - image.min()) / (image.max() - image.min())
    image *= 255
    image = image.astype(np.uint8)
    blur = cv2.GaussianBlur(image, (5,5), 0)
    _, thresh = cv2.threshold(blur, 0, 255, cv2.THRESH_BINARY+cv2.THRESH_OTSU)
    marker_internal = image > thresh
    marker_internal = segmentation.clear_border(marker_internal)
    marker_internal_labels = measure.label(marker_internal)
    areas = [r.area for r in measure.regionprops(marker_internal_labels)]
    areas.sort()
    if len(areas) > 2:
        for region in measure.regionprops(marker_internal_labels):
            if region.area < areas[-2]:
                for coordinates in region.coords:                
                       marker_internal_labels[coordinates[0], coordinates[1]] = 0
    marker_internal = marker_internal_labels > 0
    #Creation of the external Marker
    external_a = ndimage.binary_dilation(marker_internal, iterations=10)
    external_b = ndimage.binary_dilation(marker_internal, iterations=55)
    marker_external = external_b ^ external_a
    #Creation of the Watershed Marker matrix
    marker_watershed = np.zeros((512, 512), dtype=np.int)
    marker_watershed += marker_internal * 255
    marker_watershed += marker_external * 128
    return marker_internal, marker_external, marker_watershed

def seperate_lungs(image):
    #Creation of the markers as shown above:
    marker_internal, marker_external, marker_watershed = generate_markers(image)
    #Creation of the Sobel-Gradient
    sobel_filtered_dx = ndimage.sobel(image, 1)
    sobel_filtered_dy = ndimage.sobel(image, 0)
    sobel_gradient = np.hypot(sobel_filtered_dx, sobel_filtered_dy)
    sobel_gradient *= 255.0 / np.max(sobel_gradient)
    #Watershed algorithm
    watershed = morphology.watershed(sobel_gradient, marker_watershed)
    #Reducing the image created by the Watershed algorithm to its outline
    outline = ndimage.morphological_gradient(watershed, size=(3,3))
    outline = outline.astype(bool)
    #Performing Black-Tophat Morphology for reinclusion
    #Creation of the disk-kernel and increasing its size a bit
    blackhat_struct = [[0, 0, 1, 1, 1, 0, 0],
                       [0, 1, 1, 1, 1, 1, 0],
                       [1, 1, 1, 1, 1, 1, 1],
                       [1, 1, 1, 1, 1, 1, 1],
                       [1, 1, 1, 1, 1, 1, 1],
                       [0, 1, 1, 1, 1, 1, 0],
                       [0, 0, 1, 1, 1, 0, 0]]
    blackhat_struct = ndimage.iterate_structure(blackhat_struct, 8)
    #Perform the Black-Hat
    outline += ndimage.black_tophat(outline, structure=blackhat_struct)
    #Use the internal marker and the Outline that was just created to generate the lungfilter
    lungfilter = np.bitwise_or(marker_internal, outline)
    #Close holes in the lungfilter
    #fill_holes is not used here, since in some slices the heart would be reincluded by accident
    lungfilter = ndimage.morphology.binary_closing(lungfilter, structure=np.ones((5,5)), iterations=3)
    #Apply the lungfilter (note the filtered areas being assigned -2000 HU)
    segmented = np.where(lungfilter == 1, image, -2000*np.ones((512, 512)))
    return lungfilter


if __name__ == '__main__':
    path_to_save = '/datasets/bodymask_lungonly'
    image_dir = "/datasets/reference/k2_d100_st1.0"
    to_exclude = [mask.split('.h5')[0]+'.nrrd' for mask in os.listdir(path_to_save)]
    files = [file for file in os.listdir(image_dir) if (file.endswith('.nrrd') and file not in to_exclude)]
    for img_file in files:
        print('Segmenting case:', img_file)
        file_name = img_file.split('.nrrd')[0] + '.h5'
        save_path = os.path.join(path_to_save, file_name)
        lung_file = os.path.join(image_dir, img_file)
        lung_data, header = nrrd.read(lung_file)
        lung_data = lung_data.transpose((2, 1, 0))
        masks = np.stack([seperate_lungs(img) for img in lung_data])
        masks = masks.astype(np.uint8)
        hf = h5py.File(save_path, 'w')
        hf.create_dataset('data', data=masks)
        hf.close()
        print('All cases segmented!')
        




import pandas as pd
from evaluationScript.noduleCADEvaluationLUNA16 import noduleCADEvaluation
import os


if __name__ == '__main__':
    annotations_filename = '/workspace/SANet/data/ucla_filtered_testUID_70_30_split_annotGT.csv'
    val_path = '/workspace/SANet/data/ucla_filtered_testUID_70_30_split_pids.txt'
    res_path = './results_normCnd/SNGAN/k1_d100_st1/res/95/FROC/results.csv'


    noduleCADEvaluation(annotations_filename, res_path, val_path, os.path.join('./', 'tmp_res'), return_dict=False)





import logging
from collections import OrderedDict
import torch
import torch.nn as nn
from . import lr_scheduler
from . import networks
from .base_model import BaseModel
from .modules.loss import GANLoss, GradientPenaltyLoss
from tqdm import tqdm
"""
Amp allows users to easily experiment with different pure and mixed precision modes. 
Commonly-used default modes are chosen by selecting an “optimization level” or opt_level; each opt_level 
establishes a set of properties that govern Amp’s implementation of pure or mixed precision training.
- opt_level: 01 = Mixed Precision (recommended for typical use)
"""
from apex import amp
import apex
logger = logging.getLogger('base')


"""
Instantiates both generator and discriminator, put models in training, define losses, optimizers and schedulers
"""
class SRGANModel(BaseModel):
    def __init__(self, opt):
        super(SRGANModel, self).__init__(opt)
        train_opt = opt['train']
        if self.is_train:
            # define discriminator
            self.netD = networks.define_D(opt).to(self.device)  # D
            # put the generator and discriminator on training
            self.netG.train()
            self.netD.train()

        # define losses, optimizer and scheduler for training mode
        if self.is_train:
            # Set Generator pixel loss if 'pixel_weight' is greater than 0
            if train_opt['pixel_weight'] > 0:
                l_pix_type = train_opt['pixel_criterion'] # l1
                if l_pix_type == 'l1':
                    """
                    Creates a criterion that measures the mean absolute error (MAE) between 
                    each element in the input xx and target yy 
                    """
                    self.cri_pix = nn.L1Loss().to(self.device)

                elif l_pix_type == 'l2':
                    """
                    Creates a criterion that measures the mean squared error (squared L2 norm) between each
                    element in the input xx and target yy .
                    """
                    self.cri_pix = nn.MSELoss().to(self.device)

                else:
                    raise NotImplementedError('Loss type [{:s}] not recognized.'.format(l_pix_type))
                self.l_pix_w = train_opt['pixel_weight'] # get pixel weight
            else:
                logger.info('Skipping pixel loss ...')
                self.cri_pix = None

            # auxially loss
            if self.opt['network_D']['aux_lbl_loss']:
                logger.info('Including auxillary loss')
                self.aux_loss = nn.NLLLoss().to(self.device)

            # set Generator feature loss if 'feature_weight' is greater than 0
            if train_opt['feature_weight']:
                # print("feature weight is:", train_opt['feature_weight'])
                l_fea_type = train_opt['feature_criterion']
                # print("feature type:", l_fea_type)
                if l_fea_type == 'l1':
                    # print("feature criterion L1")
                    self.cri_fea = nn.L1Loss().to(self.device)
                elif l_fea_type == 'l2':
                    # print("feature criterion MSE")
                    self.cri_fea = nn.MSELoss().to(self.device)
                else:
                    raise NotImplementedError('Loss type [{:s}] not recognized.'.format(l_fea_type))
                self.l_fea_w = train_opt['feature_criterion']['feature_weight'] # get feature weight
            else:
                logger.info('Skipping feature loss ...')
                self.cri_fea = None
            
            # if feature criterion is defined, we load VGG perceptual loss
            if self.cri_fea:
                self.netF = networks.define_F(opt, use_bn=False).to(self.device)

            # print("Defining Gan Loss for real-fake image prediction")
            # print("-"*40)
            self.cri_gan = GANLoss(train_opt['gan_type'], real_label_val=1.0, fake_label_val=0.0).to(self.device)
            # print("cri_gan loss:", self.cri_gan)
            self.l_gan_w = train_opt['gan_weight']
            # print("gan weight:", self.l_gan_w)
            # D_update_ratio (how many times discriminator should be train after each generator training) 
            # and D_init_iters
            self.D_update_ratio = train_opt['D_update_ratio'] if train_opt['D_update_ratio'] else 1
            # print("D update ratio:", self.D_update_ratio)
            self.D_init_iters = train_opt['D_init_iters'] if train_opt['D_init_iters'] else 0
            # print("D init iters:", self.D_init_iters) # ?

            # if opt['gan_type'] is 'wgan', initialize gradient penality
            if "wgan" in train_opt['gan_type']:
                if train_opt['gan_type'] == "wgan-gp":
                    self.cri_gp = GradientPenaltyLoss(center=1.).to(self.device)
                elif train_opt['gan_type'] == "wgan-gp0":
                    self.cri_gp = GradientPenaltyLoss(center=0.).to(self.device)
                else:
                    raise NotImplementedError("{:s} not found".format(train_opt['gan_type']))
                self.l_gp_w = 10. # weight for gradient penality (used in 'optimize_parameters' function)

            # optimizers - weight decay for generator
            wd_G = train_opt['weight_decay_G'] if train_opt['weight_decay_G'] else 0
            optim_params = []
            """
            Returns an iterator over module parameters, yielding both the name of the parameter as well as
            the parameter itself, which we unpack into k, v.
            """
            for k, v in self.netG.named_parameters():  # optimize for a part of the model
                if v.requires_grad:
                    optim_params.append(v)
                else:
                    logger.warning('Params [{:s}] will not optimize.'.format(k))

            # initialize generator optimizer with parameters that only require gradients
            # initialize with weight decay and two beta coefficients
            self.optimizer_G = torch.optim.Adam(optim_params, lr=train_opt['lr_G'], \
                weight_decay=wd_G, betas=(train_opt['beta1_G'], train_opt['beta2_G']))

            """
            self.optimizer_G = torch.optim.RMSprop(optim_params, lr=train_opt['lr_G'], weight_decay=wd_G)
            """
            self.optimizers.append(self.optimizer_G) # add optimizer to list defined in base class
            # weight decay for discriminator
            wd_D = train_opt['weight_decay_D'] if train_opt['weight_decay_D'] else 0
            # optimizer for discriminator for entire model parameters
            self.optimizer_D = torch.optim.Adam(self.netD.parameters(), lr=train_opt['lr_D'], \
                weight_decay=wd_D, betas=(train_opt['beta1_D'], train_opt['beta2_D']))
            """
            self.optimizer_D = torch.optim.RMSprop(optim_params, lr=train_opt['lr_D'], weight_decay=wd_D)
            """
            self.optimizers.append(self.optimizer_D) # add optimizer to list defined in base class

            # configure schedulers
            if train_opt['lr_scheme'] == 'MultiStepLR':
                # for each optimizer, we have a scheduler added to a list, defined in base class
                for optimizer in self.optimizers:
                    self.schedulers.append(
                        lr_scheduler.MultiStepLR_Restart(optimizer, train_opt['lr_steps'], # [20e3, 40e3, 60e3]
                                                         restarts=train_opt['restarts'], # null
                                                         weights=train_opt['restart_weights'], # null
                                                         gamma=train_opt['lr_gamma'], # 0.5
                                                         clear_state=train_opt['clear_state'])) # None
            elif train_opt['lr_scheme'] == 'CosineAnnealingLR':
                for optimizer in self.optimizers:
                    self.schedulers.append(
                        lr_scheduler.CosineAnnealingLR_Restart(optimizer, train_opt['T_period'],
                                                               eta_min=train_opt['eta_min'],
                                                               restarts=train_opt['restarts'],
                                                               weights=train_opt['restart_weights']))
            else:
                raise NotImplementedError('Choose MultiStepLR or CosineAnnealingLR')

            # create OrderedDict
            self.log_dict = OrderedDict()
            # print("log dict:", self.log_dict)
        self.print_network(use_logger=True) # print the instantiated model
        self.load()  # load G and D if needed

    """
    Initializes model for mixed precision training, depending on 'opt_level'. Below is sample implementation
    -----------------------------------------------------------------------
    # Declare model and optimizer as usual, with default (FP32) precision
    model = torch.nn.Linear(D_in, D_out).cuda()
    optimizer = torch.optim.SGD(model.parameters(), lr=1e-3)

    # Allow Amp to perform casts as required by the opt_level
    model, optimizer = amp.initialize(model, optimizer, opt_level="O1")
    ...
    # loss.backward() becomes:
    with amp.scale_loss(loss, optimizer) as scaled_loss:
        scaled_loss.backward()
    """
    def initialize_amp(self):
        [self.netG, self.netD], [self.optimizer_G, self.optimizer_D] = \
        amp.initialize([self.netG, self.netD], [self.optimizer_G, self.optimizer_D],
                       opt_level=self.opt['opt_level'], num_losses = 2)
        if self.opt['gpu_ids']: 
            """
            Implements data parallelism at the module level. This container parallelizes 
            the application of the given module by splitting the input across the specified
            """
            assert torch.cuda.is_available()
            self.netG = nn.DataParallel(self.netG)
            self.netD = nn.DataParallel(self.netD)

    def test(self, data):
        self.netG_eval.eval()
        if self.opt['precision'] == 'fp16':
            var_L_eval = self.var_L.half()
            if self.opt['network_G']['use_attention']:
                if self.opt['network_G']['use_attention']['generate_candidate']:
                    var_L_eval_candidate = self.cand_L_test.half()
        else:
            var_L_eval = self.var_L
            if self.opt['network_G']['use_attention']:
                if self.opt['network_G']['use_attention']['generate_candidate']:
                    var_L_eval_candidate = self.cand_L_test

        opt_val = self.opt['datasets']['val']
        # HR slice num
        num_HR = int (data['LR'].size(2) * self.opt['scale'])
        # print('num HR is:', num_HR)
        # HR slice overlap
        HR_ot = int(self.opt['scale'] * self.ot)
        # print('HR_ot is:', HR_ot)
        if not opt_val['need_voxels']:
            print('---- Full size image prediction -----')
            # print('LR Var Size:', var_L_eval.shape)
            pt, H, W = self.var_L.size(2), self.var_L.size(3), self.var_L.size(4)
            pt = int(pt * self.opt['scale'])
            self.fake_H = torch.empty(1, 1,  num_HR, H, W, device=self.device)
            if self.opt['precision'] == 'fp16':
                fake_H_in_chunks = torch.empty(self.nt, 1,  pt, H, W, dtype=torch.half, device=self.device)
            else:
                fake_H_in_chunks = torch.empty(self.nt, 1,  pt, H, W, device=self.device)
            # mask, record 1 when there is value in this pixel
            stitch_mask = torch.zeros_like(self.fake_H, device=self.device)
            # print('stich mask size:', stitch_mask.shape)
            # generate predictions
            with torch.no_grad():
                if opt_val['full_volume']:
                    if self.opt['network_G']['need_embed']:
                        self.fake_H = self.netG_eval(var_L_eval, self.test_kernel_label, self.test_dose_label)
                    else:
                        self.fake_H = self.netG_eval(var_L_eval)
                else:
                    for i in range(0, self.nt):
                        if self.opt['network_G']['need_embed']:
                            fake_H_in_chunks[[i],...] = self.netG_eval(var_L_eval[[i],...], self.test_kernel_label, self.test_dose_label)
                        else:
                            # print('Chunk size: {}, going through gen size: {}'.format(i, var_L_eval[[i,...]].shape))
                            fake_H_in_chunks[[i],...] = self.netG_eval(var_L_eval[[i],...])
                    # stitch volume in z-direction
                    for i in range(0, self.nt - 1):
                        ts, te = i * (pt - HR_ot), i * (pt - HR_ot) + pt
                        self.fake_H[0, 0, ts:te, :, :] = (self.fake_H[0, 0, ts:te, :, :] * stitch_mask[0, 0, ts:te, :, :] +
                        fake_H_in_chunks[i,...].float() * (2 - stitch_mask[0, 0, ts:te, :, :])) / 2
                        stitch_mask[0, 0, ts:te, :, :] = 1.
                    # stitch last volume
                    self.fake_H[0, 0, -pt:, :, :] = \
                        (self.fake_H[0, 0, -pt:, :, :] * stitch_mask[0, 0, -pt:, :, :] +
                        fake_H_in_chunks[-1,...].float() * (2 - stitch_mask[0, 0, -pt:, :, :])) / 2
        else:
            if opt_val['need_voxels'] and not opt_val['need_voxels']['tile_x_y']:
                # print('---- Random voxel prediction-----')
                if self.opt['network_G']['need_embed']:
                    self.fake_H = self.netG_eval(var_L_eval, self.test_kernel_label, self.test_dose_label).float()
                else:
                    self.fake_H = self.netG_eval(var_L_eval).float()
            elif opt_val['need_voxels'] and opt_val['need_voxels']['tile_x_y']:
                # print('---- Sliding Tile Inference -----')
                # print('case has to be tiled back:', self.var_L.shape)
                pt, H, W = opt_val['slice_size'], data['LR'].size(3), data['LR'].size(4)
                pt = int(pt * self.opt['scale']) # 32*1 = 32
                self.fake_H = torch.empty(1, 1,  num_HR, H, W, device=self.device)
                # get predictions
                with torch.no_grad():
                    for row in range(0, data['LR'].size(3), opt_val['need_voxels']['tile_size']):
                        for column in range(0, data['LR'].size(3), opt_val['need_voxels']['tile_size']):
                            # [1, 1, 330, 64, 64]
                            LR_chunked = var_L_eval[:, :, :, row:row+opt_val['need_voxels']['tile_size'], column:column+opt_val['need_voxels']['tile_size']]
                            # get candidate if needed
                            if self.opt['network_G']['use_attention']:
                                if self.opt['network_G']['use_attention']['generate_candidate']:
                                    LR_chunked_candidate = var_L_eval_candidate[:, :, :, row:row+opt_val['need_voxels']['tile_size'], column:column+opt_val['need_voxels']['tile_size']]

                            # store chunked prediction results
                            if self.opt['precision'] == 'fp16':
                                # [12, 1, 32, 64, 64]
                                tmp_chunk_along_z = torch.empty(self.nt, 1, pt, opt_val['need_voxels']['tile_size'], opt_val['need_voxels']['tile_size'],
                                                    dtype=torch.half, device=self.device)
                            else:
                                tmp_chunk_along_z = torch.empty(self.nt, 1, pt, opt_val['need_voxels']['tile_size'], opt_val['need_voxels']['tile_size'],
                                                    device=self.device)

                            # iterate over number of blocks to get predictions
                            for i in range(0, self.nt - 1):
                                if self.opt['network_G']['need_embed']:
                                    if self.opt['network_G']['use_attention']:
                                        if self.opt['network_G']['use_attention']['generate_candidate']:
                                            tmp_chunk_along_z[i, :, :, :, :] = self.netG_eval(LR_chunked[:, :, i*(pt-self.ot):i*(pt-self.ot)+pt, :, :], 
                                                                        self.test_kernel_label, self.test_dose_label, LR_chunked_candidate[:, :, i*(pt-self.ot):i*(pt-self.ot)+pt, :, :])
                                    else:
                                        tmp_chunk_along_z[i, :, :, :, :] = self.netG_eval(LR_chunked[:, :, i*(pt-self.ot):i*(pt-self.ot)+pt, :, :], 
                                                                    self.test_kernel_label, self.test_dose_label)
                                else:
                                    tmp_chunk_along_z[i, :, :, :, :] = self.netG_eval(LR_chunked[:, :, i*(pt-self.ot):i*(pt-self.ot)+pt, :, :])
                
                            # add the last chunk
                            if self.opt['network_G']['need_embed']:
                                if self.opt['network_G']['use_attention']:
                                    if self.opt['network_G']['use_attention']['generate_candidate']:
                                        tmp_chunk_along_z[-1, :, :, :, :] = self.netG_eval(LR_chunked[:, :, -pt:, :, :], self.test_kernel_label, self.test_dose_label, LR_chunked_candidate[:, :, -pt:, :, :])
                                else:
                                    tmp_chunk_along_z[-1, :, :, :, :] = self.netG_eval(LR_chunked[:, :, -pt:, :, :], self.test_kernel_label, self.test_dose_label)
                            else:
                                tmp_chunk_along_z[-1, :, :, :, :] = self.netG_eval(LR_chunked[:, :, -pt:, :, :])

                            # reconstruct the volume along z [1, 1, 330, 64, 64]
                            reconstructed_z = torch.empty(1, 1, num_HR, opt_val['need_voxels']['tile_size'],
                                                        opt_val['need_voxels']['tile_size'], device=self.device)
                            # stitch volume along z direction
                            stitch_mask = torch.zeros_like(reconstructed_z, device=self.device)
                            for i in range(0, self.nt - 1):
                                ts, te = i * (pt - HR_ot), i * (pt - HR_ot) + pt
                                reconstructed_z[0, 0, ts:te, :, :] = (reconstructed_z[0, 0, ts:te, :, :] * stitch_mask[0, 0, ts:te, :, :] + 
                                                                        tmp_chunk_along_z[i,...].float() * (2 - stitch_mask[0, 0, ts:te, :, :])) / 2
                                stitch_mask[0, 0, ts:te, :, :] = 1.
                            # stich last volume
                            reconstructed_z[0, 0, -pt:, :, :] = \
                                (reconstructed_z[0, 0, -pt:, :, :] * stitch_mask[0, 0, -pt:, :, :] +
                                tmp_chunk_along_z[-1,...].float() * (2 - stitch_mask[0, 0, -pt:, :, :])) / 2
                            # accumulate volume together
                            self.fake_H[0, 0, :, row:row+opt_val['need_voxels']['tile_size'], column:column+opt_val['need_voxels']['tile_size']] = reconstructed_z

                            """
                            img_to_stitch = var_L_eval[volume_chunk] # [12, 1, 32, 64, 64]
                            if self.opt['precision'] == 'fp16':
                                fake_H_in_chunks = torch.empty(self.nt, 1,  pt, opt_val['need_voxels']['tile_size'], opt_val['need_voxels']['tile_size'], dtype=torch.half, device=self.device)
                                # print('fake H in chunck (empty):', fake_H_in_chunks.shape)
                            else:
                                fake_H_in_chunks = torch.empty(self.nt, 1,  pt, opt_val['need_voxels']['tile_size'], opt_val['need_voxels']['tile_size'], device=self.device)
                            
                            # get predictions
                            for i in range(0, self.nt):
                                if self.opt['network_G']['need_embed']:
                                    fake_H_in_chunks[[i],...] = self.netG_eval(img_to_stitch[[i],...], self.test_kernel_label, self.test_dose_label)
                                else:
                                    # print('Chunk size: {}, going through gen size: {}'.format(i, var_L_eval[[i,...]].shape))
                                    fake_H_in_chunks[[i],...] = self.netG_eval(img_to_stitch[[i],...])
                            # reconstruct to original volume size
                            reconstructed_z = torch.empty(1, 1, num_HR, opt_val['need_voxels']['tile_size'], opt_val['need_voxels']['tile_size'], device=self.device)
                            stitch_mask = torch.zeros_like(reconstructed_z, device=self.device)
                            # stictch prediction
                            for i in range(0, self.nt - 1):
                                ts, te = i * (pt - HR_ot), i * (pt - HR_ot) + pt
                                reconstructed_z[0, 0, ts:te, :, :] = (reconstructed_z[0, 0, ts:te, :, :] * stitch_mask[0, 0, ts:te, :, :] + 
                                                                        fake_H_in_chunks[i,...].float() * (2 - stitch_mask[0, 0, ts:te, :, :])) / 2
                                stitch_mask[0, 0, ts:te, :, :] = 1.
                            # the last volume
                            reconstructed_z[0, 0, -pt:, :, :] = \
                                (reconstructed_z[0, 0, -pt:, :, :] * stitch_mask[0, 0, -pt:, :, :] +
                                fake_H_in_chunks[-1,...].float() * (2 - stitch_mask[0, 0, -pt:, :, :])) / 2
                            # add chunk to initial volume
                            self.fake_H[0, 0, :, row:row+opt_val['need_voxels']['tile_size'], column:column+opt_val['need_voxels']['tile_size']] = reconstructed_z
                            volume_chunk += 1
                            """   
                    # print('Stitched shape after prediction:', self.fake_H.shape)             
            else:
                raise ValueError('Unknown tiling case found in cSNGAN model!')
        self.netG.train()

    """
    Feed in LR and HR data to models, calculate loss and optimize gradients
    """
    def optimize_parameters(self, step):
        # print("Parameter optimization in progress")
        # print("-"*40)
        # optimize generator
        self.optimizer_G.zero_grad() # zero-gradient
        # print("Step is:", step)
        # send 'LR' data to generator and get a 'fake_HR' images, assiginig it 'self.fake_H'
        if self.opt['network_G']['need_embed']:
            if self.opt['network_G']['use_attention']:
                if self.opt['network_G']['use_attention']['generate_candidate']:
                    self.fake_H = self.netG(self.var_L, self.train_kernel_label, self.train_dose_label, self.cand_L_train)
            else:
                self.fake_H = self.netG(self.var_L, self.train_kernel_label, self.train_dose_label)
        else:
            self.fake_H = self.netG(self.var_L)
        # print("Output after generator output:", self.fake_H.shape)
        l_g_total = 0 # loss for generator
        # print("Check for optimizing generator: {} % {} == 0 and {} > {} = {}".format(step, self.D_update_ratio, step, self.D_init_iters,
        # (step % self.D_update_ratio == 0 and step > self.D_init_iters)))
        # ------------------------ #
        # Calcualte Generator Loss #
        # ------------------------ #
        if step % self.D_update_ratio == 0 and step > self.D_init_iters:
            # print("Optimizing generator - loss will be calculated")
            # if generator pixel loss is defined, calculate pixel loss
            if self.cri_pix:
                # calculate loss between output from generator (given 'LR' data) & real 'HR'
                l_g_pix = self.cri_pix(self.fake_H, self.real_H)
                # multiply pixel_weight (1 -> all pixel are equally important) with calculate pixel loss
                # append loss to 'l_g_total' 
                l_g_total += self.l_pix_w * l_g_pix 
            
            # if feature loss (perceptual loss) is defined, calculate feature loss
            if self.cri_fea:  # feature loss
                # get real-output after passing real data ('real_H') to perceptual network (PN)
                real_fea = self.netF(self.real_H).detach()
                # get fake feature-output after passing fake HR data (self.fake_H) to PN
                fake_fea = self.netF(self.fake_H) 
                # calculate feature loss, either 'L1' or 'MSE' loss
                l_g_fea = self.cri_fea(fake_fea, real_fea)
                # multiple feature weight with feature loss and add to 'l_g_total'
                l_g_total += self.l_fea_w * l_g_fea

            # send fake 'HR', output of gan model, as input to discriminator and get fake prediction 
            if self.opt['network_D']['need_embed']:
                if self.opt['network_D']['aux_lbl_loss']:
                    pred_g_fake, pred_g_kernel_fake, pred_g_dose_fake = self.netD(self.fake_H, self.train_kernel_label, self.train_dose_label)
                else:
                    pred_g_fake = self.netD(self.fake_H, self.train_kernel_label, self.train_dose_label)
            else:
                if self.opt['network_D']['aux_lbl_loss']:
                    pred_g_fake, pred_g_kernel_fake, pred_g_dose_fake = self.netD(self.fake_H)
                else:
                    pred_g_fake = self.netD(self.fake_H)
            
            # if 'gan_type' is 'hinge', no need to call the 'self.cri_gan()' function, we can simply 
            # calculate the gan generator loss, else, utilize the 'self.cri_gan()' function
            if self.opt['train']['gan_type'] == 'hinge':
                l_g_gan = -pred_g_fake.mean()
            else:
                l_g_gan = self.cri_gan(pred_g_fake, True)
            
            # multipley gan weight ('self.l_gan_w' = 5e-3) with gan generator loss & add to 'l_g_total'
            l_g_total += self.l_gan_w * l_g_gan
            # calculate auxially loss for generator if specified
            if self.opt['network_D']['aux_lbl_loss']:
                if self.opt['network_D']['aux_lbl_loss']['apply_to_gen']:
                    g_aux_kernel_loss = self.aux_loss(pred_g_kernel_fake, self.train_kernel_label.view(-1, 1).squeeze(1))
                    g_aux_dose_loss = self.aux_loss(pred_g_dose_fake, self.train_dose_label.view(-1, 1).squeeze(1))
                    l_g_total += (g_aux_kernel_loss + g_aux_dose_loss)

            # backpropogate loss, step-up optimizer
            with amp.scale_loss(l_g_total , self.optimizer_G, loss_id=0) as errG_scaled:
                errG_scaled.backward()
            self.optimizer_G.step()
            # print("generator back-propagation done at step {}!".format(step))

        # print("Optimizing discriminator")
        # optimize discriminator
        self.optimizer_D.zero_grad()
        l_d_total = 0

        # for 'wgan-gp0', we do need gradient on real data
        if self.opt['train']['gan_type'] == 'wgan-gp0':
            # print("setting 'requires_grad' to be true for HR data")
            self.real_H.requires_grad_()

        # get prediction from discriminator based on real 'HR' data
        if self.opt['network_D']['need_embed']:
            if self.opt['network_D']['aux_lbl_loss']:
                pred_d_real, aux_kernel_real, aux_dose_real = self.netD(self.real_H, self.train_kernel_label, self.train_dose_label)
                pred_d_fake, aux_kernel_fake, aux_dose_fake = self.netD(self.fake_H.detach(), self.train_kernel_label.detach(), self.train_dose_label.detach())  # detach to avoid back propogation to G
            else:
                pred_d_real = self.netD(self.real_H, self.train_kernel_label, self.train_dose_label)
                pred_d_fake = self.netD(self.fake_H.detach(), self.train_kernel_label.detach(), self.train_dose_label.detach())  # detach to avoid back propogation to G
        else:
            if self.opt['network_D']['aux_lbl_loss']:
                pred_d_real, aux_kernel_real, aux_dose_real = self.netD(self.real_H)
                pred_d_fake, aux_kernel_fake, aux_dose_fake = self.netD(self.fake_H.detach())
            else:
                # get predictions form discriminator based on 'real_H' data
                pred_d_real = self.netD(self.real_H)
                # get predictions form discriminator based on 'fake_HR' data
                pred_d_fake = self.netD(self.fake_H.detach())  # detach to avoid back propogation to G

        # ---------------------------- #
        # Calcualte Discriminator Loss #
        # ---------------------------- #
        l_d_real = self.cri_gan(pred_d_real, True)
        l_d_fake = self.cri_gan(pred_d_fake, False)
        l_d_total = l_d_real + l_d_fake

        # calculate auxially loss for discriminator generator
        if self.opt['network_D']['aux_lbl_loss']:
            l_d_aux_kernel_real = self.aux_loss(aux_kernel_real, self.train_kernel_label.view(-1, 1).squeeze(1))
            l_d_aux_dose_real = self.aux_loss(aux_dose_real, self.train_dose_label.view(-1, 1).squeeze(1))
            l_d_aux_kernel_fake = self.aux_loss(aux_kernel_fake, self.train_kernel_label.view(-1, 1).squeeze(1))
            l_d_aux_dose_fake = self.aux_loss(aux_dose_fake, self.train_dose_label.view(-1, 1).squeeze(1))
            l_d_total += (l_d_aux_kernel_real + l_d_aux_kernel_fake + l_d_aux_dose_real + l_d_aux_dose_fake)
            # compute accuracy
            kernel_accuracy = self.compute_acc(torch.exp(aux_kernel_real), self.train_kernel_label.view(-1, 1).squeeze(1))
            dose_accuracy = self.compute_acc(torch.exp(aux_dose_real), self.train_dose_label.view(-1, 1).squeeze(1))

        # if 'wgan' in 'gan_tpye', we calculate gradient penality, multiply with gradient penalty weight and add
        # to 'l_d_gp'
        if 'wgan' in self.opt['train']['gan_type']:
            if self.opt['train']['gan_type'] == 'wgan-gp0':
                l_d_gp = self.cri_gp(self.real_H, pred_d_real)
            elif self.opt['train']['gan_type'] == 'wgan-gp':
                batch_size = self.real_H.size(0)
                eps = torch.rand(batch_size, device=self.device).view(batch_size, 1, 1, 1, 1)
                x_interp = (1 - eps) * self.real_H + eps * self.fake_H.detach()
                x_interp.requires_grad_()
                if self.opt['network_D']['need_embed']:
                    if self.opt['network_D']['aux_lbl_loss']:
                        pred_d_x_interp, prd_x_kernel_interp, pred_x_dose_interp = self.netD(x_interp, self.train_kernel_label, self.train_dose_label)
                    else:
                        pred_d_x_interp = self.netD(x_interp, self.train_kernel_label, self.train_dose_label)
                else:
                    if self.opt['network_D']['aux_lbl_loss']:
                        pred_d_x_interp, prd_x_kernel_interp, pred_x_dose_interp = self.netD(x_interp)
                    else:
                        pred_d_x_interp = self.netD(x_interp)
                l_d_gp = self.cri_gp(x_interp, pred_d_x_interp)
            else:
                raise NotImplementedError('Gan type [{:s}] not recognized'.format(self.opt['train']['gan_type']))
            l_d_total += self.l_gp_w * l_d_gp # weight for gp (self.l_gp_w) = 10

        # backpropogate loss, step-up optimizer
        with amp.scale_loss(l_d_total , self.optimizer_D, loss_id=1) as errD_scaled:
            errD_scaled.backward()
        self.optimizer_D.step()

        # print("Check for logging generator: {} % {} == 0 and {} > {} = {}".format(step, self.D_update_ratio, step, self.D_init_iters,
        # (step % self.D_update_ratio == 0 and step > self.D_init_iters)))
        
        # Set logs - Log Generator  
        if step % self.D_update_ratio == 0 and step > self.D_init_iters:
            # Generator
            # append pixel loss (L1) loss, if specified
            if self.cri_pix:
                self.log_dict['l_g_pix'] = l_g_pix.item()
            # append perceptual loss/feature loss, if specified
            if self.cri_fea:
                self.log_dict['l_g_fea'] = l_g_fea.item()
            # append gan loss
            self.log_dict['l_g_gan'] = l_g_gan.item()
            # append total loss (pixel loss (if specified) + feature loss (if specified) + gan loss/hinge loss)
            self.log_dict['l_g_total'] = l_g_total.item()

        # Log Discriminator
        self.log_dict['l_d_total'] = l_d_total.item()
        if 'wgan' in self.opt['train']['gan_type']:
            # append gradient penalty
            self.log_dict['l_d_gp'] = l_d_gp.item()
            # append wasertian distance = - (real loss + fake loss)
            self.log_dict['w_dist'] = - ( l_d_real.item() + l_d_fake.item() )# pred_d_real - pred_d_fake
        
        # D outputs (mean of output from real HR and mean of output from fake HR)
        self.log_dict['D_real'] = torch.mean(pred_d_real.detach())
        self.log_dict['D_fake'] = torch.mean(pred_d_fake.detach())
        if self.opt['network_D']['aux_lbl_loss']:
            self.log_dict['Kernel_acc'] = kernel_accuracy
            self.log_dict['Dose_acc'] = dose_accuracy
    
    """
    Returns the log dictionary that contains the gan (generator & discriminator) losses
    """
    def get_current_log(self):
        return self.log_dict

    def get_current_visuals(self, data, maskOn=True, need_HR=True):
        # print('is maskOn:', maskOn) # False during validation / True during training
        # print('need_HR:', need_HR) # True during validation and training
        out_dict = OrderedDict()
        out_dict['LR'] = self.var_L.detach()[0, 0].float() # [channel, 512, 512]
        out_dict['SR'] = self.fake_H.detach()[0, 0].float() # [channel, 512, 512]
        # print('get current visual shape LR:', out_dict['LR'].shape)
        # print('HR shape:', out_dict['SR'].shape)
        if maskOn:
            # the way we contructed mask it is 1 x 1 x depth x height x width
            mask = data['mask'].to(self.device).float()[0, 0, :]
            # print('shap of mask:', mask.shape)
            out_dict['SR'] *= mask
            # print('unique element in mask', torch.unique(mask))
        if need_HR:
            out_dict['HR'] = self.real_H.detach().float()[0, 0, :]
            if maskOn:
                out_dict['HR'] *= mask
        return out_dict

    """
    Prints out both the generator and discriminator network structure
    """
    def print_network(self, use_logger=True):
        # Generator
        s, n = self.get_network_description(self.netG)
        if isinstance(self.netG, nn.DataParallel):
            net_struc_str = '{} - {}'.format(self.netG.__class__.__name__,
                                             self.netG.module.__class__.__name__)
        else:
            net_struc_str = '{}'.format(self.netG.__class__.__name__)
        if use_logger:
            logger.info('Network G structure: {}, with parameters: {:,d}'.format(net_struc_str, n))
            logger.info(s)
        else:
            pass
            # print('Network G structure: {}, with parameters: {:,d}'.format(net_struc_str, n))
            # print(s)
        if self.is_train:
            # Discriminator
            s, n = self.get_network_description(self.netD)
            if isinstance(self.netD, nn.DataParallel):
                net_struc_str = '{} - {}'.format(self.netD.__class__.__name__,
                                                self.netD.module.__class__.__name__)
            else:
                net_struc_str = '{}'.format(self.netD.__class__.__name__)
            if use_logger:
                logger.info('Network D structure: {}, with parameters: {:,d}'.format(net_struc_str, n))
                logger.info(s)
            else:
                pass
                # print('Network D structure: {}, with parameters: {:,d}'.format(net_struc_str, n))
                # print(s)
            if self.cri_fea:  # F, Perceptual Network
                s, n = self.get_network_description(self.netF)
                if isinstance(self.netF, nn.DataParallel):
                    net_struc_str = '{} - {}'.format(self.netF.__class__.__name__,
                                                    self.netF.module.__class__.__name__)
                else:
                    net_struc_str = '{}'.format(self.netF.__class__.__name__)
                logger.info('Network F structure: {}, with parameters: {:,d}'.format(net_struc_str, n))
                logger.info(s)

    """
    Load the generator and discriminator weights if provided
    """
    def load(self, use_logger=True):
        load_path_G = self.opt['path']['pretrain_model_G']
        if load_path_G is not None: # load, model weights, if path is not None
            # 'self.load_network()' is implemented in base class
            self.load_network(load_path_G, self.netG)
            if use_logger:
                logger.info('Loading pretrained model for G [{:s}] ...'.format(load_path_G))
            else:
                print('Loading pretrained model for G [{:s}] ...'.format(load_path_G))
        load_path_D = self.opt['path']['pretrain_model_D']
        # load if opt['is_train'] is 'Train' and discriminator weight path is not None
        if self.opt['is_train'] and load_path_D is not None:
            self.load_network(load_path_D, self.netD)
            if use_logger:
                logger.info('Loading pretrained model for D [{:s}] ...'.format(load_path_D))
            else:
                print('Loading pretrained model for D [{:s}] ...'.format(load_path_D))

    """
    Save model weights, both generator and discriminator, at a particular iteration
    """
    def save(self, iter_step):
        # 'self.save_network()' is implemented in base class
        self.save_network(self.netG, 'G', iter_step)
        self.save_network(self.netD, 'D', iter_step)
    
    def run_test(self, data): 
        need_HR = False if self.opt['datasets']['dataroot_HR'] is None else True
        has_mask = False if self.opt['datasets']['maskroot_HR'] is None else True
        self.feed_test_data(data, need_HR=need_HR)
        self.test(data)  # test
        visuals = self.get_current_visuals(data, maskOn=has_mask, need_HR=need_HR)
        data['volume'] = visuals['SR'][None,None]
        return data 


import numpy as np
import torch
import os
import traceback
import time
import nrrd
import sys
import matplotlib.pyplot as plt
import logging
import argparse
import torch.nn.functional as F
from scipy.stats import norm
from torch.utils.data import DataLoader
from tqdm import tqdm
from torch.autograd import Variable
from torch.nn.parallel.data_parallel import data_parallel
from scipy.ndimage.measurements import label
from scipy.ndimage import center_of_mass
from net.sanet import SANet
from dataset.collate import train_collate, test_collate, eval_collate
from dataset.bbox_reader import BboxReader
from config import config
import pandas as pd
from evaluationScript.noduleCADEvaluationLUNA16 import noduleCADEvaluation

this_module = sys.modules[__name__]
os.environ['CUDA_VISIBLE_DEVICES'] = '2'


parser = argparse.ArgumentParser()
parser.add_argument('--net', '-m', metavar='NET', default=config['net'],
                    help='neural net')
parser.add_argument("--mode", type=str, default = 'eval',
                    help="you want to test or val")
parser.add_argument("--weight", type=str, default='./results/model/model.ckpt',
                    help="path to model weights to be used")
parser.add_argument("--dicom-path", type=str, default=None,
                    help="path to dicom files of patient")
parser.add_argument("--out-dir", type=str, default=config['out_dir'],
                    help="path to save the results")
parser.add_argument("--test-set-name", type=str, default=config['test_set_name'],
                    help="path to save the results")


def main():
    logging.basicConfig(format='[%(levelname)s][%(asctime)s] %(message)s', level=logging.INFO)
    args = parser.parse_args()

    if args.mode == 'eval':
        data_dir = config['preprocessed_data_dir']
        test_set_name = args.test_set_name
        num_workers = 16
        initial_checkpoint = args.weight
        net = args.net
        out_dir = args.out_dir

        net = getattr(this_module, net)(config)
        net = net.cuda()

        if initial_checkpoint:
            print('[Loading model from %s]' % initial_checkpoint)
            checkpoint = torch.load(initial_checkpoint)
            epoch = checkpoint['epoch']
            net.load_state_dict(checkpoint['state_dict'])
        else:
            print('No model weight file specified')
            return

        print('out_dir', out_dir)
        save_dir = os.path.join(out_dir, 'res', str(epoch))
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        if not os.path.exists(os.path.join(save_dir, 'FROC')):
            os.makedirs(os.path.join(save_dir, 'FROC'))

        dataset = BboxReader(data_dir, test_set_name, config, mode='eval')
        test_loader = DataLoader(dataset, batch_size=1, shuffle=False,
                                 num_workers=num_workers, pin_memory=False, collate_fn=train_collate)
        eval(net, test_loader, save_dir)
    else:
        logging.error('Mode %s is not supported' % (args.mode))


def eval(net, dataset, save_dir=None):
    net.set_mode('eval')
    net.use_rcnn = True

    print('Total # of eval data %d' % (len(dataset)))
    for i, (input, truth_bboxes, truth_labels) in enumerate(dataset):
        try:
            input = Variable(input).cuda()
            truth_bboxes = np.array(truth_bboxes)
            truth_labels = np.array(truth_labels)
            pid = dataset.dataset.filenames[i]

            print('[%d] Predicting %s' % (i, pid))

            with torch.no_grad():
                net.forward(input, truth_bboxes, truth_labels)
            detections = net.rpn_proposals.cpu().numpy()
            if len(detections):
                detections = detections[:, 1:-1]
                np.save(os.path.join(save_dir, '%s_detections.npy' % (pid)), detections)

            # Clear gpu memory
            del input, truth_bboxes, truth_labels
            torch.cuda.empty_cache()

        except Exception as e:
            del input, truth_bboxes, truth_labels
            torch.cuda.empty_cache()
            traceback.print_exc()

            return
    
    # Generate prediction csv for the use of performing FROC analysis
    res = []
    for pid in dataset.dataset.filenames:
        if os.path.exists(os.path.join(save_dir, '%s_detections.npy' % (pid))):
            detections = np.load(os.path.join(save_dir, '%s_detections.npy' % (pid)))
            detections = detections[:, [3, 2, 1, 4, 0]]
            names = np.array([[pid]] * len(detections))
            res.append(np.concatenate([names, detections], axis=1))
    
    res = np.concatenate(res, axis=0)
    col_names = ['pid','center_x','center_y','center_z','diameter', 'probability']
    eval_dir = os.path.join(save_dir, 'FROC')
    res_path = os.path.join(eval_dir, 'results.csv')
    
    df = pd.DataFrame(res, columns=col_names)
    df.to_csv(res_path, index=False)

    # Start evaluating
    if not os.path.exists(os.path.join(eval_dir, 'res')):
        os.makedirs(os.path.join(eval_dir, 'res'))

    annotations_filename = './data/luna16_annot_center.csv'
    # test_anno_center.csv is a csv file with 'center_x','center_y','center_z', and 'diameter'. You can obtain these parameters by using 'xmin', 'xmax', etc.
    val_path = './data/luna16_pid.txt'
    noduleCADEvaluation(annotations_filename, res_path, val_path, os.path.join(eval_dir, 'res'))


if __name__ == '__main__':
    main()
    # python test.py --weight='./model.ckpt' --out_dir='./results/' --test_set_name='./data/luna16_pid.txt'



import math
import random
import numpy as np
import torch
import cv2

####################
# image processing
# process on numpy image
####################

"""
Converts incoming h5 data input (uint16 - D,H,W) to torch tensor float in a given range range [0, 1500].
For dataloader use, inplace operation
"""
class ImgToTensor(object):
    def __call__(self, sample, clip=False, raw_data_range=1500.):
        if clip:
            img = torch.from_numpy(np.clip(np.array(sample, np.float32, copy=False), 0, raw_data_range))
        else:
            img = torch.from_numpy(np.array(sample, np.float32, copy=False))
        return img.float().div_(raw_data_range)  # range is from 0 to 1500 for png

def augment(img_list, hflip=True, rot=True):
    # horizontal flip OR rotate
    hflip = hflip and random.random() < 0.5
    vflip = rot and random.random() < 0.5
    rot90 = rot and random.random() < 0.5

    def _augment(img):
        if hflip: img = img[:, ::-1, :]
        if vflip: img = img[::-1, :, :]
        if rot90: img = img.transpose(1, 0, 2)
        return img

    return [_augment(img) for img in img_list]


def channel_convert(in_c, tar_type, img_list):
    # conversion among BGR, gray and y
    if in_c == 3 and tar_type == 'gray':  # BGR to gray
        gray_list = [cv2.cvtColor(img, cv2.COLOR_BGR2GRAY) for img in img_list]
        return [np.expand_dims(img, axis=2) for img in gray_list]
    elif in_c == 3 and tar_type == 'y':  # BGR to y
        y_list = [bgr2ycbcr(img, only_y=True) for img in img_list]
        return [np.expand_dims(img, axis=2) for img in y_list]
    elif in_c == 1 and tar_type == 'RGB':  # gray/y to BGR
        return [cv2.cvtColor(img, cv2.COLOR_GRAY2BGR) for img in img_list]
    else:
        return img_list


def rgb2ycbcr(img, only_y=True):
    '''same as matlab rgb2ycbcr
    only_y: only return Y channel
    Input:
        uint8, [0, 255]
        float, [0, 1]
    '''
    in_img_type = img.dtype
    img.astype(np.float32)
    if in_img_type != np.uint8:
        img *= 255.
    # convert
    if only_y:
        rlt = np.dot(img, [65.481, 128.553, 24.966]) / 255.0 + 16.0
    else:
        rlt = np.matmul(img, [[65.481, -37.797, 112.0], [128.553, -74.203, -93.786],
                              [24.966, 112.0, -18.214]]) / 255.0 + [16, 128, 128]
    if in_img_type == np.uint8:
        rlt = rlt.round()
    else:
        rlt /= 255.
    return rlt.astype(in_img_type)


def bgr2ycbcr(img, only_y=True):
    '''bgr version of rgb2ycbcr
    only_y: only return Y channel
    Input:
        uint8, [0, 255]
        float, [0, 1]
    '''
    in_img_type = img.dtype
    img.astype(np.float32)
    if in_img_type != np.uint8:
        img *= 255.
    # convert
    if only_y:
        rlt = np.dot(img, [24.966, 128.553, 65.481]) / 255.0 + 16.0
    else:
        rlt = np.matmul(img, [[24.966, 112.0, -18.214], [128.553, -74.203, -93.786],
                              [65.481, -37.797, 112.0]]) / 255.0 + [16, 128, 128]
    if in_img_type == np.uint8:
        rlt = rlt.round()
    else:
        rlt /= 255.
    return rlt.astype(in_img_type)


def ycbcr2rgb(img):
    '''same as matlab ycbcr2rgb
    Input:
        uint8, [0, 255]
        float, [0, 1]
    '''
    in_img_type = img.dtype
    img.astype(np.float32)
    if in_img_type != np.uint8:
        img *= 255.
    # convert
    rlt = np.matmul(img, [[0.00456621, 0.00456621, 0.00456621], [0, -0.00153632, 0.00791071],
                          [0.00625893, -0.00318811, 0]]) * 255.0 + [-222.921, 135.576, -276.836]
    if in_img_type == np.uint8:
        rlt = rlt.round()
    else:
        rlt /= 255.
    return rlt.astype(in_img_type)


def modcrop(img_in, scale):
    # img_in: Numpy, HWC or HW
    img = np.copy(img_in)
    if img.ndim == 2:
        H, W = img.shape
        H_r, W_r = H % scale, W % scale
        img = img[:H - H_r, :W - W_r]
    elif img.ndim == 3:
        H, W, C = img.shape
        H_r, W_r = H % scale, W % scale
        img = img[:H - H_r, :W - W_r, :]
    else:
        raise ValueError('Wrong img ndim: [{:d}].'.format(img.ndim))
    return img


####################
# Functions
####################
# matlab 'imresize' function, now only support 'bicubic'
def cubic(x):
    absx = torch.abs(x)
    absx2 = absx**2
    absx3 = absx**3
    return (1.5*absx3 - 2.5*absx2 + 1) * ((absx <= 1).type_as(absx)) + \
        (-0.5*absx3 + 2.5*absx2 - 4*absx + 2) * (((absx > 1)*(absx <= 2)).type_as(absx))


def calculate_weights_indices(in_length, out_length, scale, kernel, kernel_width, antialiasing):
    if (scale < 1) and (antialiasing):
        # Use a modified kernel to simultaneously interpolate and antialias- larger kernel width
        kernel_width = kernel_width / scale

    # Output-space coordinates
    x = torch.linspace(1, out_length, out_length)

    # Input-space coordinates. Calculate the inverse mapping such that 0.5
    # in output space maps to 0.5 in input space, and 0.5+scale in output
    # space maps to 1.5 in input space.
    u = x / scale + 0.5 * (1 - 1 / scale)

    # What is the left-most pixel that can be involved in the computation?
    left = torch.floor(u - kernel_width / 2)

    # What is the maximum number of pixels that can be involved in the
    # computation?  Note: it's OK to use an extra pixel here; if the
    # corresponding weights are all zero, it will be eliminated at the end
    # of this function.
    P = math.ceil(kernel_width) + 2

    # The indices of the input pixels involved in computing the k-th output
    # pixel are in row k of the indices matrix.
    indices = left.view(out_length, 1).expand(out_length, P) + torch.linspace(0, P - 1, P).view(
        1, P).expand(out_length, P)

    # The weights used to compute the k-th output pixel are in row k of the
    # weights matrix.
    distance_to_center = u.view(out_length, 1).expand(out_length, P) - indices
    # apply cubic kernel
    if (scale < 1) and (antialiasing):
        weights = scale * cubic(distance_to_center * scale)
    else:
        weights = cubic(distance_to_center)
    # Normalize the weights matrix so that each row sums to 1.
    weights_sum = torch.sum(weights, 1).view(out_length, 1)
    weights = weights / weights_sum.expand(out_length, P)

    # If a column in weights is all zero, get rid of it. only consider the first and last column.
    weights_zero_tmp = torch.sum((weights == 0), 0)
    if not math.isclose(weights_zero_tmp[0], 0, rel_tol=1e-6):
        indices = indices.narrow(1, 1, P - 2)
        weights = weights.narrow(1, 1, P - 2)
    if not math.isclose(weights_zero_tmp[-1], 0, rel_tol=1e-6):
        indices = indices.narrow(1, 0, P - 2)
        weights = weights.narrow(1, 0, P - 2)
    weights = weights.contiguous()
    indices = indices.contiguous()
    sym_len_s = -indices.min() + 1
    sym_len_e = indices.max() - in_length
    indices = indices + sym_len_s - 1
    return weights, indices, int(sym_len_s), int(sym_len_e)


def imresize(img, scale, antialiasing=True):
    # Now the scale should be the same for H and W
    # input: img: CHW RGB [0,1]
    # output: CHW RGB [0,1] w/o round

    in_C, in_H, in_W = img.size()
    out_C, out_H, out_W = in_C, math.ceil(in_H * scale), math.ceil(in_W * scale)
    kernel_width = 4
    kernel = 'cubic'

    # Return the desired dimension order for performing the resize.  The
    # strategy is to perform the resize first along the dimension with the
    # smallest scale factor.
    # Now we do not support this.

    # get weights and indices
    weights_H, indices_H, sym_len_Hs, sym_len_He = calculate_weights_indices(
        in_H, out_H, scale, kernel, kernel_width, antialiasing)
    weights_W, indices_W, sym_len_Ws, sym_len_We = calculate_weights_indices(
        in_W, out_W, scale, kernel, kernel_width, antialiasing)
    # process H dimension
    # symmetric copying
    img_aug = torch.FloatTensor(in_C, in_H + sym_len_Hs + sym_len_He, in_W)
    img_aug.narrow(1, sym_len_Hs, in_H).copy_(img)

    sym_patch = img[:, :sym_len_Hs, :]
    inv_idx = torch.arange(sym_patch.size(1) - 1, -1, -1).long()
    sym_patch_inv = sym_patch.index_select(1, inv_idx)
    img_aug.narrow(1, 0, sym_len_Hs).copy_(sym_patch_inv)

    sym_patch = img[:, -sym_len_He:, :]
    inv_idx = torch.arange(sym_patch.size(1) - 1, -1, -1).long()
    sym_patch_inv = sym_patch.index_select(1, inv_idx)
    img_aug.narrow(1, sym_len_Hs + in_H, sym_len_He).copy_(sym_patch_inv)

    out_1 = torch.FloatTensor(in_C, out_H, in_W)
    kernel_width = weights_H.size(1)
    for i in range(out_H):
        idx = int(indices_H[i][0])
        out_1[0, i, :] = img_aug[0, idx:idx + kernel_width, :].transpose(0, 1).mv(weights_H[i])
        out_1[1, i, :] = img_aug[1, idx:idx + kernel_width, :].transpose(0, 1).mv(weights_H[i])
        out_1[2, i, :] = img_aug[2, idx:idx + kernel_width, :].transpose(0, 1).mv(weights_H[i])

    # process W dimension
    # symmetric copying
    out_1_aug = torch.FloatTensor(in_C, out_H, in_W + sym_len_Ws + sym_len_We)
    out_1_aug.narrow(2, sym_len_Ws, in_W).copy_(out_1)

    sym_patch = out_1[:, :, :sym_len_Ws]
    inv_idx = torch.arange(sym_patch.size(2) - 1, -1, -1).long()
    sym_patch_inv = sym_patch.index_select(2, inv_idx)
    out_1_aug.narrow(2, 0, sym_len_Ws).copy_(sym_patch_inv)

    sym_patch = out_1[:, :, -sym_len_We:]
    inv_idx = torch.arange(sym_patch.size(2) - 1, -1, -1).long()
    sym_patch_inv = sym_patch.index_select(2, inv_idx)
    out_1_aug.narrow(2, sym_len_Ws + in_W, sym_len_We).copy_(sym_patch_inv)

    out_2 = torch.FloatTensor(in_C, out_H, out_W)
    kernel_width = weights_W.size(1)
    for i in range(out_W):
        idx = int(indices_W[i][0])
        out_2[0, :, i] = out_1_aug[0, :, idx:idx + kernel_width].mv(weights_W[i])
        out_2[1, :, i] = out_1_aug[1, :, idx:idx + kernel_width].mv(weights_W[i])
        out_2[2, :, i] = out_1_aug[2, :, idx:idx + kernel_width].mv(weights_W[i])

    return out_2


def imresize_np(img, scale, antialiasing=True):
    # Now the scale should be the same for H and W
    # input: img: Numpy, HWC BGR [0,1]
    # output: HWC BGR [0,1] w/o round
    img = torch.from_numpy(img)

    in_H, in_W, in_C = img.size()
    out_C, out_H, out_W = in_C, math.ceil(in_H * scale), math.ceil(in_W * scale)
    kernel_width = 4
    kernel = 'cubic'

    # Return the desired dimension order for performing the resize.  The
    # strategy is to perform the resize first along the dimension with the
    # smallest scale factor.
    # Now we do not support this.

    # get weights and indices
    weights_H, indices_H, sym_len_Hs, sym_len_He = calculate_weights_indices(
        in_H, out_H, scale, kernel, kernel_width, antialiasing)
    weights_W, indices_W, sym_len_Ws, sym_len_We = calculate_weights_indices(
        in_W, out_W, scale, kernel, kernel_width, antialiasing)
    # process H dimension
    # symmetric copying
    img_aug = torch.FloatTensor(in_H + sym_len_Hs + sym_len_He, in_W, in_C)
    img_aug.narrow(0, sym_len_Hs, in_H).copy_(img)

    sym_patch = img[:sym_len_Hs, :, :]
    inv_idx = torch.arange(sym_patch.size(0) - 1, -1, -1).long()
    sym_patch_inv = sym_patch.index_select(0, inv_idx)
    img_aug.narrow(0, 0, sym_len_Hs).copy_(sym_patch_inv)

    sym_patch = img[-sym_len_He:, :, :]
    inv_idx = torch.arange(sym_patch.size(0) - 1, -1, -1).long()
    sym_patch_inv = sym_patch.index_select(0, inv_idx)
    img_aug.narrow(0, sym_len_Hs + in_H, sym_len_He).copy_(sym_patch_inv)

    out_1 = torch.FloatTensor(out_H, in_W, in_C)
    kernel_width = weights_H.size(1)
    for i in range(out_H):
        idx = int(indices_H[i][0])
        out_1[i, :, 0] = img_aug[idx:idx + kernel_width, :, 0].transpose(0, 1).mv(weights_H[i])
        out_1[i, :, 1] = img_aug[idx:idx + kernel_width, :, 1].transpose(0, 1).mv(weights_H[i])
        out_1[i, :, 2] = img_aug[idx:idx + kernel_width, :, 2].transpose(0, 1).mv(weights_H[i])

    # process W dimension
    # symmetric copying
    out_1_aug = torch.FloatTensor(out_H, in_W + sym_len_Ws + sym_len_We, in_C)
    out_1_aug.narrow(1, sym_len_Ws, in_W).copy_(out_1)

    sym_patch = out_1[:, :sym_len_Ws, :]
    inv_idx = torch.arange(sym_patch.size(1) - 1, -1, -1).long()
    sym_patch_inv = sym_patch.index_select(1, inv_idx)
    out_1_aug.narrow(1, 0, sym_len_Ws).copy_(sym_patch_inv)

    sym_patch = out_1[:, -sym_len_We:, :]
    inv_idx = torch.arange(sym_patch.size(1) - 1, -1, -1).long()
    sym_patch_inv = sym_patch.index_select(1, inv_idx)
    out_1_aug.narrow(1, sym_len_Ws + in_W, sym_len_We).copy_(sym_patch_inv)

    out_2 = torch.FloatTensor(out_H, out_W, in_C)
    kernel_width = weights_W.size(1)
    for i in range(out_W):
        idx = int(indices_W[i][0])
        out_2[:, i, 0] = out_1_aug[:, idx:idx + kernel_width, 0].mv(weights_W[i])
        out_2[:, i, 1] = out_1_aug[:, idx:idx + kernel_width, 1].mv(weights_W[i])
        out_2[:, i, 2] = out_1_aug[:, idx:idx + kernel_width, 2].mv(weights_W[i])

    return out_2.numpy()


